from mathics.core.symbols import Symbol

# DirectedEdge and UndirectedEdge are process in the parser of mathics-core
# as symbols in System`.
SymbolDirectedEdge = Symbol("System`DirectedEdge")
SymbolGraph = Symbol("Pymathics`Graph")
SymbolGraphBox = Symbol("Pymathics`GraphBox")
SymbolUndirectedEdge = Symbol("System`UndirectedEdge")
SymbolTwoWayRule = Symbol("System`TwoWayRule")



import networkx as nx
from mathics.core.symbols import SymbolConstant, SymbolFalse, SymbolTrue

from pymathics.graph.base import Graph


def eval_TreeGraphQ(g: Graph) -> SymbolConstant:
    """
    Returns SymbolTrue if g is a (networkx) tree and SymbolFalse
    otherwise.
    """
    if not isinstance(g, Graph):
        return SymbolFalse
    return SymbolTrue if nx.is_tree(g.G) else SymbolFalse



"""
Curated Graphs
"""

import networkx as nx
from typing import Callable, Dict, Optional, Tuple
from mathics.core.evaluation import Evaluation

from pymathics.graph.base import Graph, _NetworkXBuiltin, graph_helper


class GraphData(_NetworkXBuiltin):
    """
    <url>
    :WMA link:https://reference.wolfram.com/language/ref/GraphData.html</url>

    <dl>
      <dt>'GraphData[$name$]'
      <dd>Returns a graph with the specified name.
    </dl>

    >> GraphData["PappusGraph"]
     = -Graph-
    """

    summary_text = "create a graph by name"

    def eval(
        self, name, expression, evaluation: Evaluation, options: dict
    ) -> Optional[Graph]:
        "GraphData[name_String, OptionsPattern[GraphData]]"
        py_name = name.get_string_value()
        fn, layout = WL_TO_NETWORKX_FN.get(py_name, (None, None))
        if not fn:
            if not py_name.endswith("_graph"):
                py_name += "_graph"
            if py_name in ("LCF_graph", "make_small_graph"):
                # These graphs require parameters
                return
            import inspect

            fn = dict(inspect.getmembers(nx, inspect.isfunction)).get(py_name, None)
            # parameters = inspect.signature(nx.diamond_graph).parameters.values()
            # if len([p for p in list(parameters) if p.kind in [inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD]]) != 0:
            #     return
        if fn:
            g = graph_helper(fn, options, False, layout, evaluation)
            if g is not None:
                g.G.name = py_name
            return g


WL_TO_NETWORKX_FN: Dict[str, Tuple[Callable, Optional[str]]] = {
    "DodecahedralGraph": (nx.dodecahedral_graph, None),
    "DiamondGraph": (nx.diamond_graph, "spring"),
    "PappusGraph": (nx.pappus_graph, "circular"),
    "IsohedralGraph": (nx.icosahedral_graph, "spring"),
    "PetersenGraph": (nx.petersen_graph, None),
}

# TODO: ExampleData



# -*- coding: utf-8 -*-

"""
Centralities

<url>:Centralities:https://en.wikipedia.org/wiki/Centrality</url>


Routines to evaluate centralities of a graph.

In graph theory and network analysis, the centrality is a \
ranking between pairs of node according some metric.

"""

import networkx as nx
from mathics.core.atoms import Integer, Real
from mathics.core.convert.expression import ListExpression

from pymathics.graph.base import _NetworkXBuiltin


def _components(G):
    if G.is_directed():
        return nx.strongly_connected_components(G)
    else:
        return nx.connected_components(G)


class _Centrality(_NetworkXBuiltin):
    options = {
        "WorkingPrecision": "MachinePrecision",
    }
    pass


class _ComponentwiseCentrality(_Centrality):
    def _centrality(self, g, weight):
        raise NotImplementedError

    def _compute(self, graph, evaluation, reverse=False, normalized=True, **kwargs):
        vertices = graph.vertices
        G, weight = graph.coalesced_graph(evaluation)
        if reverse:
            G = G.reverse()

        components = list(_components(G))
        components = [c for c in components if len(c) > 1]

        result = [0] * len(vertices)
        for bunch in components:
            g = G.subgraph(bunch)
            centrality = self._centrality(g, weight, **kwargs)
            values = [centrality.get(v, 0) for v in vertices]
            if normalized:
                s = sum(values) * len(components)
            else:
                s = 1
            if s > 0:
                for i, x in enumerate(values):
                    result[i] += x / s

        return ListExpression(*[Real(x) for x in result])


class BetweennessCentrality(_Centrality):
    """
    <url>
    :Betweenness centrality:
    https://en.wikipedia.org/wiki/Betweenness_centrality</url> (<url>
    :NetworkX:
    https://networkx.org/documentation/networkx-2.8.8/reference/algorithms/generated/networkx.algorithms.centrality.betweenness_centrality.html</url>,\
    <url>
    :WMA:
    https://reference.wolfram.com/language/ref/BetweennessCentrality.html</url>)

    <dl>
      <dt>'BetweennessCentrality'[$g$]
      <dd>gives a list of betweenness centralities for the vertices \
          in a 'Graph' or a list of edges $g$.
    </dl>

    >> g = Graph[{a -> b, b -> c, d -> c, d -> a, e -> c, d -> b}]
     = -Graph-

    >> BetweennessCentrality[g]
     = {0., 1., 0., 0., 0.}

    >> g = Graph[{a -> b, b -> c, c -> d, d -> e, e -> c, e -> a}]
     = -Graph-

    >> BetweennessCentrality[g]
     = ...
    """

    summary_text = "get Betweenness centrality"

    def eval(self, graph, expression, evaluation, options):
        "%(name)s[graph_, OptionsPattern[%(name)s]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if graph:
            weight = graph.update_weights(evaluation)
            centrality = nx.betweenness_centrality(
                graph.G, normalized=False, weight=weight
            )
            return ListExpression(
                *[Real(centrality.get(v, 0.0)) for v in graph.vertices],
            )


class ClosenessCentrality(_Centrality):
    """
    <url>
    :Betweenness centrality:
    https://en.wikipedia.org/wiki/Closeness_centrality</url> (<url>
    :NetworkX:
    https://networkx.org/documentation/networkx-2.8.8/reference/algorithms/generated/\
    networkx.algorithms.centrality.closeness_centrality.html</url>,\
    <url>
    :WMA:
    https://reference.wolfram.com/language/ref/ClosenessCentrality.html</url>)


    <dl>
      <dt>'ClosenessCentrality'[$g$]
      <dd>gives a list of closeness centralities for the vertices \
          in a 'Graph' or a list of edges $g$.
    </dl>

     >> g = Graph[{a -> b, b -> c, d -> c, d -> a, e -> c, d -> b}]
      = -Graph-

     >> ClosenessCentrality[g]
      = {0.666667, 1., 0., 1., 1.}

     >> g = Graph[{a -> b, b -> c, c -> d, d -> e, e -> c, e -> a}]
      = -Graph-

    >> ClosenessCentrality[g]
      = {0.4, 0.4, 0.4, 0.5, 0.666667}
    """

    summary_text = "get the closeness centrality"

    def eval(self, graph, expression, evaluation, options):
        "%(name)s[graph_, OptionsPattern[%(name)s]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if graph:
            weight = graph.update_weights(evaluation)
            G = graph.G
            if G.is_directed():
                G = G.reverse()
            centrality = nx.closeness_centrality(G, distance=weight, wf_improved=False)
            return ListExpression(
                *[Real(centrality.get(v, 0.0)) for v in graph.vertices],
            )


class DegreeCentrality(_Centrality):
    """
    <url>
    :Degree centrality:
    https://en.wikipedia.org/wiki/Degree_centrality</url> (<url>
    :NetworkX:
    https://networkx.org/documentation/networkx-2.8.8/reference/algorithms/generated/\
    networkx.algorithms.centrality.degree_centrality.html</url>,\
    <url>
    :WMA:
    https://reference.wolfram.com/language/ref/DegreeCentrality.html</url>)

    <dl>
      <dt>'DegreeCentrality'[$g$]
      <dd>gives a list of degree centralities for the vertices \
          in a 'Graph' or a list of edges $g$.
    </dl>

    >> g = Graph[{a -> b, b <-> c, d -> c, d -> a, e <-> c, d -> b}]
     = -Graph-

    >> DegreeCentrality[g]
     = ...

    >> DegreeCentrality[g, "In"]
     = ...

    >> DegreeCentrality[g, "Out"]
     = ...
    """

    summary_text = "get the degree centrality"

    def _from_dict(self, graph, centrality):
        s = len(graph.G) - 1  # undo networkx's normalization
        return ListExpression(
            *[Integer(s * centrality.get(v, 0)) for v in graph.vertices],
        )

    def eval(self, graph, expression, evaluation, options):
        "Pymathics`DegreeCentrality[graph_, OptionsPattern[]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if graph:
            return self._from_dict(graph, nx.degree_centrality(graph.G))

    def eval_in(self, graph, expression, evaluation, options):
        '%(name)s[graph_, "In", OptionsPattern[]]'
        graph = self._build_graph(graph, evaluation, options, expression)
        if graph:
            return self._from_dict(graph, nx.in_degree_centrality(graph.G))

    def eval_out(self, graph, expression, evaluation, options):
        '%(name)s[graph_, "Out", OptionsPattern[]]'
        graph = self._build_graph(graph, evaluation, options, expression)
        if graph:
            return self._from_dict(graph, nx.out_degree_centrality(graph.G))


class EigenvectorCentrality(_ComponentwiseCentrality):
    """
    <url>
    :Eigenvector Centrality:
    https://en.wikipedia.org/wiki/Eigenvector_centrality</url> (<url>
    :NetworkX:
    https://networkx.org/documentation/networkx-2.8.8/reference/algorithms\
/generated/networkx.algorithms.centrality.eigenvector_centrality.html</url>,\
<url>
    :WMA:
    https://reference.wolfram.com/language/ref/EgenvectorCentrality.html</url>)

    <dl>
      <dt>'EigenvectorCentrality'[$g$]
      <dd>gives a list of eigenvector centralities for\
          the vertices in the graph $g$.
      <dt>'EigenvectorCentrality'[$g$, "In"]
      <dd>gives a list of eigenvector in-centralities for\
          the vertices in the graph $g$.
      <dt>'EigenvectorCentrality'[$g$, "Out"]
      <dd>gives a list of eigenvector out-centralities for\
          the vertices in the graph $g$.
    </dl>

    >> g = Graph[{a -> b, b -> c, c -> d, d -> e, e -> c, e -> a}]; EigenvectorCentrality[g, "In"]
     = {0.16238, 0.136013, 0.276307, 0.23144, 0.193859}

    >> EigenvectorCentrality[g, "Out"]
     = {0.136013, 0.16238, 0.193859, 0.23144, 0.276307}

    >> g = Graph[{a <-> b, b <-> c, c <-> d, d <-> e, e <-> c, e <-> a}]; EigenvectorCentrality[g]
     = {0.162435, 0.162435, 0.240597, 0.193937, 0.240597}

    >> g = Graph[{a <-> b, b <-> c, a <-> c, d <-> e, e <-> f, f <-> d, e <-> d}]; EigenvectorCentrality[g]
     = {0.166667, 0.166667, 0.166667, 0.183013, 0.183013, 0.133975}

    #> g = Graph[{a -> b, b -> c, c -> d, b -> e, a -> e}]; EigenvectorCentrality[g]
     = {0., 0., 0., 0., 0.}

    >> g = Graph[{a -> b, b -> c, c -> d, b -> e, a -> e, c -> a}]; EigenvectorCentrality[g]
     = {0.333333, 0.333333, 0.333333, 0., 0.}
    """

    summary_text = "compute Eigenvector centralities"

    def _centrality(self, g, weight):
        return nx.eigenvector_centrality(g, max_iter=10000, tol=1.0e-7, weight=weight)

    def eval(self, graph, expression, evaluation, options):
        "%(name)s[graph_, OptionsPattern[%(name)s]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if graph:
            return self._compute(graph, evaluation)

    def eval_in_out(self, graph, dir, expression, evaluation, options):
        "%(name)s[graph_, dir_String, OptionsPattern[%(name)s]]"
        py_dir = dir.get_string_value()
        if py_dir not in ("In", "Out"):
            return
        graph = self._build_graph(graph, evaluation, options, expression)
        if graph:
            return self._compute(graph, evaluation, py_dir == "Out")


class HITSCentrality(_Centrality):
    """
    <url>
    :NetworkX:
    https://networkx.org/documentation/networkx-2.8.8/reference/algorithms/\
generated/networkx.algorithms.link_analysis.hits_alg.hits.html</url>, \
    <url>
    :WMA:
    https://reference.wolfram.com/language/ref/HITSCentrality.html</url>

    <dl>
      <dt>'HITSCentrality'[$g$]
      <dd>gives a list of authority and hub centralities for\
          the vertices in the graph $g$.
    </dl>

    """

    summary_text = "get HITS centrality"

    def eval(self, graph, expression, evaluation, options):
        "%(name)s[graph_, OptionsPattern[%(name)s]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if graph:
            G, _ = graph.coalesced_graph(evaluation)  # FIXME warn if weight > 1

            tol = 1.0e-14
            _, a = nx.hits(G, normalized=True, tol=tol)
            h, _ = nx.hits(G, normalized=False, tol=tol)

            def _crop(x):
                return 0 if x < tol else x

            vertices = graph.vertices
            return ListExpression(
                ListExpression(*[Real(_crop(a.get(v, 0))) for v in vertices]),
                ListExpression(*[Real(_crop(h.get(v, 0))) for v in vertices]),
            )


class KatzCentrality(_ComponentwiseCentrality):
    """
    <url>
    :Katz Centrality:
    https://en.wikipedia.org/wiki/Katz_centrality</url> (<url>
    :NetworkX:
    https://networkx.org/documentation/networkx-2.8.8/reference/algorithms\
/generated/networkx.algorithms.centrality.katz_centrality.html\
#networkx.algorithms.centrality.katz_centrality</url>, \
    <url>
    :WMA:
    https://reference.wolfram.com/language/ref/KatzCentrality.html</url>)

    <dl>
      <dt>'KatzCentrality'[$g$, $alpha$]
      <dd>gives a list of Katz centralities for the \
          vertices in the graph $g$ and weight $alpha$.
      <dt>'KatzCentrality'[$g$, $alpha$, $beta$]
      <dd>gives a list of Katz centralities for the \
          vertices in the graph $g$ and weight $alpha$ and initial centralities $beta$.
    </dl>

    >> g = Graph[{a -> b, b -> c, c -> d, d -> e, e -> c, e -> a}]
     = -Graph-
    >> KatzCentrality[g, 0.2]
     = {1.25202, 1.2504, 1.5021, 1.30042, 1.26008}

    >> g = Graph[{a <-> b, b <-> c, a <-> c, d <-> e, e <-> f, f <-> d, e <-> d}]
     = -Graph-

    >> KatzCentrality[g, 0.1]
     = {1.25, 1.25, 1.25, 1.41026, 1.41026, 1.28205}
    """

    summary_text = "get the Katz centrality"

    rules = {
        "Pymathics`KatzCentrality[Pymathics`g_, Pymathics`alpha_]": "Pymathics`KatzCentrality[Pymathics`g, Pymathics`alpha, 1]",
    }

    def _centrality(self, g, weight, alpha, beta):
        return nx.katz_centrality(
            g, alpha=alpha, beta=beta, normalized=False, weight=weight
        )

    def eval(self, graph, alpha, beta, expression, evaluation, options):
        "Pymathics`KatzCentrality[Pymathics`graph_, alpha_, beta_, OptionsPattern[%(name)s]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if graph:
            try:
                py_alpha = alpha.to_mpmath()
                py_beta = beta.to_mpmath()
            except NotImplementedError:
                return
            if py_alpha is None or py_beta is None:
                return
            return self._compute(
                graph, evaluation, normalized=False, alpha=py_alpha, beta=py_beta
            )


class PageRankCentrality(_Centrality):
    """
    <url>
    :Pagerank Centrality:
    https://en.wikipedia.org/wiki/Pagerank</url> (<url>
    :NetworkX:
    https://networkx.org/documentation/networkx-2.8.8/reference/algorithms\
    /generated/networkx.algorithms.link_analysis.pagerank_alg.pagerank.html</url>,\
    <url>
    :WMA:
    https://reference.wolfram.com/language/ref/PageRankCentrality.html</url>)

    <dl>
      <dt>'PageRankCentrality'[$g$, $alpha$]
      <dd>gives a list of page rank centralities for the \
          vertices in the graph $g$ and weight $alpha$.
      <dt>'PageRankCentrality'[$g$, $alpha$, $beta$]
      <dd>gives a list of page rank centralities for the \
          vertices in the graph $g$ and weight $alpha$ and initial centralities $beta$.
    </dl>

    >> g = Graph[{a -> d, b -> c, d -> c, d -> a, e -> c, d -> c}]; PageRankCentrality[g, 0.2]
     = {0.184502, 0.207565, 0.170664, 0.266605, 0.170664}
    """

    summary_text = "get the page rank centralities"

    def eval_alpha_beta(self, graph, alpha, expression, evaluation, options):
        "%(name)s[graph_, alpha_, OptionsPattern[%(name)s]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if graph:
            py_alpha = float(alpha.to_mpmath())
            if py_alpha is None:
                return
            G, weight = graph.coalesced_graph(evaluation)
            centrality = nx.pagerank(G, alpha=py_alpha, weight=weight, tol=1.0e-7)
            return ListExpression(
                *[Real(centrality.get(v, 0)) for v in graph.vertices],
            )



"""
evaluation methods for Parametric Graphs
"""

from typing import Optional

import networkx as nx
from mathics.core.atoms import Integer
from mathics.core.evaluation import Evaluation

from pymathics.graph.base import Graph, graph_helper
from pymathics.graph.eval.harary import hkn_harary_graph


def eval_complete_graph(
    self, n: Integer, expression, evaluation: Evaluation, options: dict
) -> Optional[Graph]:
    py_n = n.value

    if py_n < 1:
        evaluation.message(self.get_name(), "ilsmp", expression)
        return

    args = (py_n,)
    g = graph_helper(
        nx.complete_graph, options, False, "circular", evaluation, None, *args
    )
    if not g:
        return None

    g.G.n = n
    return g


def eval_full_rary_tree(
    self, r: Integer, n: Integer, expression, evaluation: Evaluation, options: dict
) -> Optional[Graph]:
    """
    Call ``networkx.full_rary_tree()`` using parameters, ``r`` and ``t``.
    """
    py_r = r.value

    if py_r < 0:
        evaluation.message(self.get_name(), "ilsmp", expression)
        return

    py_n = n.value
    if py_n < 0:
        evaluation.message(self.get_name(), "ilsmp", expression)
        return

    args = (py_r, py_n)
    g = graph_helper(nx.full_rary_tree, options, True, "tree", evaluation, 0, *args)
    if not g:
        return None

    g.G.r = r
    g.G.n = n
    return g


def eval_hkn_harary(
    self, k: Integer, n: Integer, expression, evaluation: Evaluation, options: dict
) -> Optional[Graph]:
    py_k = k.value

    if py_k < 0:
        evaluation.message(self.get_name(), "ilsmp", expression)
        return

    py_n = n.value
    if py_n < 0:
        evaluation.message(self.get_name(), "ilsmp2", expression)
        return

    args = (py_k, py_n)
    g = graph_helper(
        hkn_harary_graph, options, False, "circular", evaluation, None, *args
    )
    if not g:
        return None
    g.k = py_k
    return g



# -*- coding: utf-8 -*-
"""
Parametric Graphs
"""

from typing import Optional

import networkx as nx
from mathics.core.atoms import Integer, Integer2
from mathics.core.evaluation import Evaluation
from mathics.core.expression import Expression

from pymathics.graph.base import (
    Graph,
    SymbolUndirectedEdge,
    _graph_from_list,
    _NetworkXBuiltin,
    graph_helper,
)
from pymathics.graph.eval.harary import hnm_harary_graph
from pymathics.graph.eval.parametric import (
    eval_complete_graph,
    eval_full_rary_tree,
    eval_hkn_harary,
)
from pymathics.graph.tree import DEFAULT_TREE_OPTIONS

# TODO: Can this code can be DRY'd more?


class BalancedTree(_NetworkXBuiltin):
    """
    <url>
    :WMA:
    https://reference.wolfram.com/language/ref/BalancedTree.html</url>

    <dl>
      <dt>'BalancedTree[$r$, $h$]'
      <dd>Returns the perfectly balanced $r$-ary tree of height $h$.

      In this tree produced, all non-leaf nodes will have $r$ children and \
      the height of the path from root $r$ to any leaf will be $h$.
    </dl>

    >> BalancedTree[2, 3]
     = -Graph-

    """

    messages = {
        "ilsmp": "Expected a non-negative integer at position 1 in ``.",
        "ilsmp2": "Expected a non-negative integer at position 2 in ``.",
        "mem": "Out of memory",
    }

    options = DEFAULT_TREE_OPTIONS
    summary_text = "build a balanced tree graph"

    def eval(
        self, r: Integer, h: Integer, expression, evaluation: Evaluation, options: dict
    ) -> Optional[Graph]:
        "%(name)s[r_Integer, h_Integer, OptionsPattern[%(name)s]]"
        py_r = r.value

        if py_r < 0:
            evaluation.message(self.get_name(), "ilsmp", expression)
            return None

        py_h = h.value
        if py_h < 0:
            evaluation.message(self.get_name(), "ilsmp2", expression)
            return None

        args = (py_r, py_h)
        g = graph_helper(nx.balanced_tree, options, True, "tree", evaluation, 0, *args)
        if not g:
            return None
        g.G.r = r
        g.G.h = h
        return g


class BarbellGraph(_NetworkXBuiltin):
    """
    <url>
    :Barbell graph:https://en.wikipedia.org/wiki/Barbell_graph </url> (<url>
    :NetworkX:https://networkx.org/documentation/networkx-2.8.8/reference/\
generated/networkx.generators.classic.barbell_graph.html</url>, <url>
    :Wolfram MathWorld:
    https://mathworld.wolfram.com/BarbellGraph.html</url>)

    <dl>
      <dt>'BarbellGraph[$m1$, $m2$]'
      <dd>Barbell Graph: two complete graphs connected by a path.
    </dl>

    >> BarbellGraph[4, 1]
     = -Graph-

    """

    messages = {
        "ilsmp": "Expected a non-negative integer at position 1 in ``.",
        "ilsmp2": "Expected a non-negative integer at position 2 in ``.",
    }
    summary_text = "build a n-m Barbell graph"

    def eval(
        self,
        m1: Integer,
        m2: Integer,
        expression,
        evaluation: Evaluation,
        options: dict,
    ) -> Optional[Graph]:
        "BarbellGraph[m1_Integer, m2_Integer, OptionsPattern[BarbellGraph]]"
        py_m1 = m1.value

        if py_m1 < 0:
            evaluation.message(self.get_name(), "ilsmp", expression)
            return

        py_m2 = m2.value
        if py_m2 < 0:
            evaluation.message(self.get_name(), "ilsmp2", expression)
            return

        args = (py_m1, py_m2)
        g = graph_helper(
            nx.barbell_graph, options, False, "spring", evaluation, None, *args
        )
        if not g:
            return None

        g.G.m1 = py_m1
        g.G.m2 = py_m2
        return g


class BinomialTree(_NetworkXBuiltin):
    """
    <url>
    :Binomial tree:
    https://en.wikipedia.org/wiki/Binomial_heap</url> (<url>
    :NetworkX:
    https://networkx.org/documentation/networkx-2.8.8/reference/\
generated/networkx.generators.classic.binomial_tree.html</url>, <url>
    :WMA:https://reference.wolfram.com/language/ref/BinomialTree.html</url>)

    <dl>
      <dt>'BinomialTree[$n$]'
      <dd>Returns the Binomial Tree of order $n$.

      The binomial tree of order $n$ with root $R$ is defined as:

      If $k$=0, 'B[$k$]' = 'B[0]' = {$R$}. i.e., the binomial tree of order \
      zero consists of a single node, $R$.

      If $k$>0, 'B[$k$]' = {$R$, 'B[0]', 'B[1]' ... 'B[$k$]'}, i.e., the binomial tree \
      of order $k$>0 comprises the root $R$, and $k$ binomial subtrees, \
      'B[0]' to 'B[$k$]'.

      Binomial trees are the underlying data structure in <url>
      :Binomial heaps:
      https://en.wikipedia.org/wiki/Binomial_heap#Binomial_tree</url>.
    </dl>

    >> BinomialTree[3]
     = -Graph-

    """

    messages = {
        "ilsmp": "Expected a non-negative integer at position 1 in ``.",
        "mem": "Out of memory",
    }
    summary_text = "build a binomial tree"

    def eval(
        self, n: Integer, expression, evaluation: Evaluation, options: dict
    ) -> Optional[Graph]:
        "%(name)s[n_Integer, OptionsPattern[%(name)s]]"
        py_n = n.value

        if py_n < 0:
            evaluation.message(self.get_name(), "ilsmp", expression)
            return

        args = (py_n,)
        g = graph_helper(nx.binomial_tree, options, True, "tree", evaluation, 0, *args)
        if not g:
            return None
        g.G.n = n
        return g


class CompleteGraph(_NetworkXBuiltin):
    """
    <url>
    :Complete Multipartite Graph:
    https://en.wikipedia.org/wiki/Multipartite_graph</url> (<url>
    :NetworkX:
    https://networkx.org/documentation/networkx-2.8.8/reference/\
generated/networkx.generators.classic.complete_multipartite_graph.html</url>, <url>
    :WMA:
    https://reference.wolfram.com/language/ref/CompleteGraph.html</url>)

    <dl>
      <dt>'CompleteGraph[$n$]'
      <dd>Returns the complete graph with $n$ vertices, $K_n$.
    </dl>

    >> CompleteGraph[8]
     = -Graph-
    """

    messages = {
        "ilsmp": "Expected a positive integer at position 1 in ``.",
    }

    summary_text = "build a completely-connected graph"

    def eval(self, n: Integer, expression, evaluation: Evaluation, options: dict):
        "CompleteGraph[n_Integer, OptionsPattern[CompleteGraph]]"
        return eval_complete_graph(self, n, expression, evaluation, options)

    def eval_multipartite(
        self, n, evaluation: Evaluation, options: dict
    ) -> Optional[Graph]:
        "CompleteGraph[n_List, OptionsPattern[CompleteGraph]]"
        if all(isinstance(i, Integer) for i in n.elements):
            return Graph(nx.complete_multipartite_graph(*[i.value for i in n.elements]))


class CompleteKaryTree(_NetworkXBuiltin):
    """
    <url>
    :M-ary Tree:
    https://en.wikipedia.org/wiki/M-ary_tree</url> (<url>
    :NetworkX:
    https://networkx.org/documentation/networkx-2.8.8/reference/\
generated/networkx.generators.classic.full_rary_tree.html</url>, <url>
    :WMA:
    https://reference.wolfram.com/language/ref/CompleteKaryTree.html</url>)

    <dl>
      <dt>'CompleteKaryTree[$n$, $k$]'
      <dd>Creates a complete $k$-ary tree of $n$ levels.
    </dl>

    In the returned tree, with $n$ nodes, the from root $R$ to any \
    leaf be $k$.

    >> CompleteKaryTree[2, 3]
     = -Graph-

    >> CompleteKaryTree[3]
     = -Graph-

    """

    options = DEFAULT_TREE_OPTIONS
    summary_text = "build a complete k-ary tree"

    def eval(
        self, k: Integer, n: Integer, expression, evaluation: Evaluation, options: dict
    ):
        "CompleteKaryTree[n_Integer, k_Integer, OptionsPattern[CompleteKaryTree]]"

        n_int = n.value
        k_int = k.value

        new_n_int = int(((k_int**n_int) - 1) / (k_int - 1))
        return eval_full_rary_tree(
            self, k, Integer(new_n_int), expression, evaluation, options
        )

    # FIXME: can be done with rules?
    def eval_2(self, n: Integer, expression, evaluation: Evaluation, options: dict):
        "CompleteKaryTree[n_Integer, OptionsPattern[CompleteKaryTree]]"

        n_int = n.value

        new_n_int = int(2**n_int) - 1
        return eval_full_rary_tree(
            self, Integer2, Integer(new_n_int), expression, evaluation, options
        )


class CycleGraph(_NetworkXBuiltin):
    """
    <url>:Cycle Graph:
    https://en.wikipedia.org/wiki/Cycle_graph</url> (<url>
    :WMA:
    https://reference.wolfram.com/language/ref/CycleGraph.html</url>)

    <dl>
      <dt>'CycleGraph[$n$]'
      <dd>Returns the cycle graph with $n$ vertices $C_n$.
    </dl>

    >> CycleGraph[5, PlotLabel -> "C_i"]
     = -Graph-
    """

    summary_text = "build a cycle graph"

    def eval(
        self, n: Integer, expression, evaluation: Evaluation, options: dict
    ) -> Optional[Graph]:
        "CycleGraph[n_Integer, OptionsPattern[CycleGraph]]"
        n_int = n.value
        if n_int < 3:
            return eval_complete_graph(self, n, expression, evaluation, options)
        else:
            return eval_hkn_harary(self, Integer2, n, expression, evaluation, options)


class GraphAtlas(_NetworkXBuiltin):
    """
    <url>:NetworkX:
    https://networkx.org/documentation/networkx-2.8.8/reference/\
generated/networkx.generators.atlas.graph_atlas.html
    </url>

    <dl>
      <dt>'GraphAtlas[$n$]'
      <dd>Returns graph number $i$ from the NetworkX's Graph \
      Atlas. There are about 1200 of them and get large as $i$ \
      increases.
    </dl>

    >> GraphAtlas[1000]
     = -Graph-
    """

    messages = {
        "ilsmp": "Expected a positive integer at position 1 in ``.",
    }
    summary_text = "retrieve a graph by number from the NetworkX Atlas"

    def eval(
        self, n: Integer, expression, evaluation: Evaluation, options: dict
    ) -> Optional[Graph]:
        "GraphAtlas[n_Integer, OptionsPattern[GraphAtlas]]"
        py_n = n.value

        if py_n < 1:
            evaluation.message(self.get_name(), "ilsmp", expression)
            return

        args = (py_n,)
        g = graph_helper(
            nx.graph_atlas, options, False, "spring", evaluation, None, *args
        )
        if not g:
            return None
        g.n = py_n
        return g


class HknHararyGraph(_NetworkXBuiltin):
    """
    <url>:NetworkX:
    https://networkx.org/documentation/networkx-2.8.8/reference\
/generated/networkx.generators.harary_graph.hkn_harary_graph.html#hkn-harary-graph</url>, <url>
    :WMA:
    https://reference.wolfram.com/language/ref/HknHararyGraph.html</url>

    <dl>
      <dt>'HknHararyGraph[$k$, $n$]'
      <dd>Returns the Harary graph with given node connectivity and node number.

      This second generator gives the Harary graph that minimizes the \
      number of edges in the graph with given node connectivity and   \
      number of nodes.

      Harary, F.  The Maximum Connectivity of a Graph.  \
      Proc. Nat. Acad. Sci. USA 48, 1142-1146, 1962.
    </dl>

    >> HknHararyGraph[3, 10]
     = -Graph-

    """

    messages = {
        "ilsmp": "Expected a non-negative integer at position 1 in ``.",
        "ilsmp2": "Expected a non-negative integer at position 2 in ``.",
    }
    summary_text = "build a Hkn Harary graph"

    def eval(self, k, n, expression, evaluation: Evaluation, options: dict):
        "%(name)s[k_Integer, n_Integer, OptionsPattern[%(name)s]]"
        return eval_hkn_harary(self, k, n, expression, evaluation, options)


class HmnHararyGraph(_NetworkXBuiltin):
    """
    <url>
    :NetworkX:
    https://networkx.org/documentation/networkx-2.8.8/reference\
/generated/networkx.generators.harary_graph.hnm_harary_graph.html</url>, <url>
   :WMA:
    https://reference.wolfram.com/language/ref/HmnHararyGraph.html</url>

    <dl>
      <dt>'HmnHararyGraph[$m$, $n$]'
      <dd>Returns the Harary graph with given numbers of nodes and edges.

      This generator gives the Harary graph that maximizes the node \
      connectivity with given number of nodes and given number of \
      edges.

      Harary, F.  The Maximum Connectivity of a Graph.\
      Proc. Nat. Acad. Sci. USA 48, 1142-1146, 1962.
    </dl>

    >> HmnHararyGraph[5, 10]
     = -Graph-
    """

    messages = {
        "ilsmp": "Expected a non-negative integer at position 1 in ``.",
        "ilsmp2": "Expected a non-negative integer at position 2 in ``.",
    }

    summary_text = "build a Hmn Harary graph"

    def eval(
        self, n: Integer, m: Integer, expression, evaluation: Evaluation, options: dict
    ) -> Optional[Graph]:
        "%(name)s[n_Integer, m_Integer, OptionsPattern[%(name)s]]"
        py_n = n.value

        if py_n < 0:
            evaluation.message(self.get_name(), "ilsmp", expression)
            return

        py_m = m.value

        if py_m < 0:
            evaluation.message(self.get_name(), "ilsmp2", expression)
            return

        args = (py_n, py_m)
        g = graph_helper(
            hnm_harary_graph, options, False, "circular", evaluation, None, *args
        )
        if not g:
            return None
        g.m = py_m
        return g


class KaryTree(_NetworkXBuiltin):
    """
    <url>
    :M-ary Tree:https://en.wikipedia.org/wiki/M-ary_tree
    </url>


    <dl>
      <dt>'KaryTree[$r$, $n$]'
      <dd>Creates binary tree of $n$ vertices.
    </dl>

    <dl>
      <dt>'KaryTree[$n$, $k$]'
      <dd>Creates $k$-ary tree with $n$ vertices.
    </dl>

    >> KaryTree[10]
     = -Graph-

    >> KaryTree[3, 10]
     = -Graph-

    """

    messages = {
        "ilsmp": "Expected a non-negative integer at position 1 in ``.",
        "ilsmp2": "Expected a non-negative integer at position 2 in ``.",
        "mem": "Out of memory",
    }

    options = DEFAULT_TREE_OPTIONS
    summary_text = "build a k-ary tree"

    def eval(
        self, n: Integer, expression, evaluation: Evaluation, options: dict
    ) -> Optional[Graph]:
        "KaryTree[n_Integer, OptionsPattern[KaryTree]]"
        return eval_full_rary_tree(self, Integer2, n, expression, evaluation, options)

    def eval_with_k(
        self, n: Integer, k: Integer, expression, evaluation: Evaluation, options: dict
    ) -> Optional[Graph]:
        "KaryTree[n_Integer, k_Integer, OptionsPattern[KaryTree]]"
        return eval_full_rary_tree(self, k, n, expression, evaluation, options)


class LadderGraph(_NetworkXBuiltin):
    """
    <url>
    :Ladder graph:https://en.wikipedia.org/wiki/Ladder_graph</url> (<url>
    :NetworkX:
    https://networkx.org/documentation/networkx-2.8.8/reference\
/generated/networkx.generators.classic.ladder_graph.html</url>)

    <dl>
      <dt>'LadderGraph[$n$]'
      <dd>Returns the Ladder graph of length $n$.
    </dl>

    >> LadderGraph[8]
     = -Graph-
    """

    messages = {
        "ilsmp": "Expected a positive integer at position 1 in ``.",
    }
    summary_text = "build a ladder tree"

    def eval(
        self, n: Integer, expression, evaluation: Evaluation, options: dict
    ) -> Optional[Graph]:
        "LadderGraph[n_Integer, OptionsPattern[LadderGraph]]"
        py_n = n.value

        if py_n < 1:
            evaluation.message(self.get_name(), "ilsmp", expression)
            return

        args = (py_n,)
        g = graph_helper(
            nx.ladder_graph, options, False, "spring", evaluation, 0, *args
        )
        if not g:
            return None
        g.G.n = n
        return g


class PathGraph(_NetworkXBuiltin):
    """
    <url>
    :Path graph:https://en.wikipedia.org/wiki/Path_graph
    </url> (<url>:WMA:https://reference.wolfram.com/language/ref/PathGraph.html
    </url>)
    <dl>
      <dt>'PathGraph[{$v_1$, $v_2$, ...}]'
      <dd>Returns a Graph with a path with vertices $v_i$ and \
      edges between $v-i$ and $v_i+1$ .
    </dl>

    >> PathGraph[{1, 2, 3}]
     = -Graph-
    """

    summary_text = "build a path graph"

    def eval(self, e, evaluation: Evaluation, options: dict) -> Graph:
        "PathGraph[e_List, OptionsPattern[PathGraph]]"
        elements = e.elements

        def edges():
            for u, v in zip(elements, elements[1:]):
                yield Expression(SymbolUndirectedEdge, u, v)

        g = _graph_from_list(edges(), options)
        g.G.graph_layout = (
            options["System`GraphLayout"].get_string_value() or "spiral_equidistant"
        )
        return g


class RandomTree(_NetworkXBuiltin):
    """
    <url>:NetworkX:
    https://networkx.org/documentation/networkx-2.8.8/reference\
/generated/networkx.generators.trees.random_tree.html</url>, <url>
    :WMA:
    https://reference.wolfram.com/language/ref/RandomTree.html</url>

    <dl>
      <dt>'RandomTree[$n$]'
      <dd>Returns a uniformly random tree on $n$ nodes.
    </dl>

    >> RandomTree[3]
     = -Graph-

    """

    messages = {
        "ilsmp": "Expected a non-negative integer at position 1 in ``.",
    }

    summary_text = "build a random tree"

    def eval(
        self, n: Integer, expression, evaluation: Evaluation, options: dict
    ) -> Optional[Graph]:
        "RandomTree[n_Integer, OptionsPattern[RandomTree]]"
        py_n = n.value

        if py_n < 0:
            evaluation.message(self.get_name(), "ilsmp", expression)
            return

        args = (py_n,)
        g = graph_helper(nx.random_tree, options, False, "tree", evaluation, 0, *args)
        if not g:
            return None
        g.G.n = n
        return g


class StarGraph(_NetworkXBuiltin):
    """
    <url>
    :Star graph:https://en.wikipedia.org/wiki/Star_graph
    </url>(<url>:NetworkX:
    https://networkx.org/documentation/networkx-2.8.8/reference\
/generated/networkx.generators.classic.star_graph.html
    </url>, <url>:WMA:
    https://reference.wolfram.com/language/ref/StarGraph.html
    </url>)
    <dl>
      <dt>'StarGraph[$n$]'
      <dd>Returns a star graph with $n$ vertices.
    </dl>

    >> StarGraph[8]
     = -Graph-
    """

    messages = {
        "ilsmp": "Expected a positive integer at position 1 in ``.",
    }
    summary_text = "build a star graph"

    def eval(
        self, n: Integer, expression, evaluation: Evaluation, options: dict
    ) -> Optional[Graph]:
        "StarGraph[n_Integer, OptionsPattern[StarGraph]]"
        py_n = n.value

        if py_n < 1:
            evaluation.message(self.get_name(), "ilsmp", expression)
            return

        args = (py_n,)
        g = graph_helper(nx.star_graph, options, False, "spring", evaluation, 0, *args)
        if not g:
            return None
        g.G.n = py_n
        return g



"""
Random Graphs
"""

import networkx as nx
from mathics.builtin.numbers.randomnumbers import RandomEnv
from mathics.core.atoms import Integer, Integer1
from mathics.core.evaluation import Evaluation
from mathics.core.list import ListExpression

from pymathics.graph.base import (
    Graph,
    _convert_networkx_graph,
    _NetworkXBuiltin,
    _process_graph_options,
    has_directed_option,
)


class RandomGraph(_NetworkXBuiltin):
    """
    <url>
    :WMA link:https://reference.wolfram.com/language/ref/RandomGraph.html
    </url>

    <dl>
      <dt>'RandomGraph[{$n$, $m$}]'
      <dd>Returns a pseudorandom graph with $n$ vertices and $m$ edges.

      <dt>'RandomGraph[{$n$, $m$}, $k$]'
      <dd>Returns list of $k$ RandomGraph[{$n$, $m$}].
    </dl>
    """

    summary_text = "build a random graph"

    def _generate(
        self, n: Integer, m: Integer, k: Integer, evaluation: Evaluation, options: dict
    ) -> Graph:
        py_n = n.value
        py_m = m.value
        py_k = k.value
        is_directed = has_directed_option(options)

        with RandomEnv(evaluation) as _:
            for _ in range(py_k):
                # seed = rand.randint(0, 2 ** 63 - 1) # 2**63 is too large
                G = nx.gnm_random_graph(py_n, py_m, directed=is_directed)
                yield _convert_networkx_graph(G, options)

    def eval_nm(
        self, n: Integer, m: Integer, expression, evaluation: Evaluation, options: dict
    ) -> Graph:
        "RandomGraph[{n_Integer, m_Integer}, OptionsPattern[RandomGraph]]"
        g = list(self._generate(n, m, Integer1, evaluation, options))[0]
        _process_graph_options(g, options)
        return g

    def eval_nmk(
        self,
        n: Integer,
        m: Integer,
        k: Integer,
        expression,
        evaluation: Evaluation,
        options: dict,
    ) -> Graph:
        "RandomGraph[{n_Integer, m_Integer}, k_Integer, OptionsPattern[RandomGraph]]"
        return ListExpression(*self._generate(n, m, k, evaluation, options))



# -*- coding: utf-8 -*-

"""
Core routines for working with Graphs.
"""

# uses networkx

from collections import defaultdict
from inspect import isgenerator
from typing import Callable, Optional, Union

from mathics.core.builtin import AtomBuiltin, Builtin
from mathics.core.atoms import Atom, Integer, Integer0, Integer1, Integer2, String
from mathics.core.convert.expression import ListExpression, from_python, to_mathics_list
from mathics.core.element import BaseElement
from mathics.core.expression import Expression
from mathics.core.pattern import pattern_objects
from mathics.core.symbols import Symbol, SymbolList, SymbolTrue
from mathics.core.systemsymbols import (
    SymbolBlank,
    SymbolCases,
    SymbolFailed,
    SymbolMissing,
    SymbolRGBColor,
    SymbolRule,
)
from mathics.eval.patterns import Matcher

from pymathics.graph.boxes import GraphBox
from pymathics.graph.collections import _EdgeCollection
from pymathics.graph.graphsymbols import (
    SymbolDirectedEdge,
    SymbolGraph,
    SymbolTwoWayRule,
    SymbolUndirectedEdge,
)


WL_MARKER_TO_NETWORKX = {
    "Circle": "o",
    "Diamond": "D",
    "Square": "s",
    "Star": "*",
    "Pentagon": "p",
    "Octagon": "8",
    "Hexagon": "h",
    "Triangle": "^",
    # And many others. Is there a list somewhere?
}

WL_COLOR_TO_NETWORKX = {
    "Green": "g",
    "Blue": "b",
    # And many others. Is there a list somewhere?
}

WL_LAYOUT_TO_NETWORKX = {
    "CircularEmbedding": "circular",
    "SpiralEmbedding": "spiral",
    "SpectralEmbedding": "spectral",
    "SpringEmbedding": "spring",
    # And many others. Is there a list somewhere?
}

DEFAULT_GRAPH_OPTIONS = {
    "DirectedEdges": "False",
    "EdgeStyle": "{}",
    "EdgeWeight": "{}",
    "GraphLayout": "Null",
    "PlotLabel": "Null",
    "VertexLabels": "False",
    "VertexSize": "{}",
    "VertexShape": '"Circle"',
    "VertexStyle": "{}",
}

import networkx as nx


def graph_helper(
    graph_generator_func: Callable,
    options: dict,
    can_digraph: bool,
    graph_layout: Optional[str],
    evaluation,
    root: Optional[int] = None,
    *args,
    **kwargs,
) -> Optional["Graph"]:
    """
    This is a wrapping function for a NetworkX function of some sort indicated in ``graph_generator``,
    e.g. ``nx.complete_graph``.

    ``args`` and ``kwargs`` are passed to the NetworkX function, while ``can_digraph`` determines
    whether ``create_using=nx.DiGraph``is added to the NetworkX graph-generation function call.

    Parameter ``options`` and ``graph_layout`` are Mathics3 Graph
    options; ``graph_layout`` has a ``String()`` wrapped around it
    ``root`` is used when the graph is a tree.
    """
    should_digraph = can_digraph and has_directed_option(options)
    try:
        G = (
            graph_generator_func(*args, create_using=nx.DiGraph, **kwargs)
            if should_digraph
            else graph_generator_func(*args, **kwargs)
        )
    except MemoryError:
        evaluation.message("Graph", "mem", evaluation)
        return None
    if (
        graph_layout and not options["System`GraphLayout"].get_string_value()
        if "System`GraphLayout" in options
        else False
    ):
        options["System`GraphLayout"] = String(graph_layout)

    g = Graph(G)
    _process_graph_options(g, options)

    if root is not None:
        G.root = g.root = root
    return g


def has_directed_option(options: dict) -> bool:
    return options.get("System`DirectedEdges", False)


def _process_graph_options(g, options: dict) -> None:
    """
    Handle common graph-related options like VertexLabels, PlotLabel, VertexShape, etc.
    """
    # FIXME: for now we are adding both to both g and g.G.
    # g is where it is used in format. However we should wrap this as our object.
    # Access in G which might be better, currently isn't used.
    g.G.vertex_labels = g.vertex_labels = (
        options["System`VertexLabels"].to_python()
        if "System`VertexLabels" in options
        else False
    )
    shape = (
        options["System`VertexShape"].get_string_value()
        if "System`VertexShape" in options
        else "Circle"
    )

    g.G.node_shape = g.node_shape = WL_MARKER_TO_NETWORKX.get(shape, shape)

    color = (
        options["System`VertexStyle"].get_string_value()
        if "System`VertexStyle" in options
        else "Blue"
    )

    g.graph_layout = (
        options["System`GraphLayout"].get_string_value()
        if "System`GraphLayout" in options
        else ""
    )

    g.G.graph_layout = g.graph_layout = WL_LAYOUT_TO_NETWORKX.get(
        g.graph_layout, g.graph_layout
    )

    g.G.node_color = g.node_color = WL_COLOR_TO_NETWORKX.get(color, color)

    g.G.title = g.title = (
        options["System`PlotLabel"].get_string_value()
        if "System`PlotLabel" in options
        else None
    )


def _circular_layout(G):
    return nx.drawing.circular_layout(G, scale=1)


def _spectral_layout(G):
    return nx.drawing.spectral_layout(G, scale=2)


def _shell_layout(G):
    return nx.drawing.shell_layout(G, scale=2)


def _generic_layout(G, warn):
    return nx.nx_pydot.graphviz_layout(G, prog="dot")


def _path_layout(G, root):
    v = root
    x = 0
    y = 0

    k = 0
    d = 0

    pos = {}
    neighbors = G.neighbors(v)

    for _ in range(len(G)):
        pos[v] = (x, y)

        if not neighbors:
            break
        try:
            v = next(neighbors) if isgenerator(neighbors) else neighbors[0]
        except StopIteration:
            break
        neighbors = G.neighbors(v)

        if k == 0:
            if d < 1 or neighbors:
                d += 1
            x += d
        elif k == 1:
            y += d
        elif k == 2:
            if neighbors:
                d += 1
            x -= d
        elif k == 3:
            y -= d

        k = (k + 1) % 4

    return pos


def _auto_layout(G, warn):
    path_root = None

    for v, d in G.degree(G.nodes):
        if d == 1 and G.neighbors(v):
            path_root = v
        elif d > 2:
            path_root = None
            break

    if path_root is not None:
        return _path_layout(G, path_root)
    else:
        return _generic_layout(G, warn)


def _convert_networkx_graph(G, options):
    mapping = dict((v, Integer(i)) for i, v in enumerate(G.nodes))
    G = nx.relabel_nodes(G, mapping)
    [Expression(SymbolUndirectedEdge, u, v) for u, v in G.edges]
    return Graph(
        G,
        **options,
    )


_default_minimum_distance = 0.3


def _vertex_style(expr):
    return expr


def _edge_style(expr):
    return expr


def _parse_property(expr, attr_dict=None):
    if expr.has_form("Rule", 2):
        name, value = expr.elements
        if isinstance(name, Symbol):
            if attr_dict is None:
                attr_dict = {}
            attr_dict[name.get_name()] = value
    elif expr.has_form("List", None):
        for element in expr.elements:
            attr_dict = _parse_property(element, attr_dict)
    return attr_dict


class _NetworkXBuiltin(Builtin):
    requires = ("networkx",)

    options = DEFAULT_GRAPH_OPTIONS

    messages = {
        "graph": "Expected a graph at position 1 in ``.",
        "inv": "The `1` at position `2` in `3` does not belong to the graph at position 1.",
    }

    def _not_a_vertex(self, expression, pos, evaluation):
        evaluation.message(self.get_name(), "inv", "vertex", pos, expression)

    def _not_an_edge(self, expression, pos, evaluation):
        evaluation.message(self.get_name(), "inv", "edge", pos, expression)

    def _build_graph(
        self,
        graph: Union["Graph", Expression],
        evaluation,
        options: dict,
        expr,
        quiet=False,
    ) -> Optional["Graph"]:
        head = graph.get_head()
        if head is SymbolGraph and isinstance(graph, Atom) and hasattr(graph, "G"):
            return graph
        elif head is SymbolList:
            return _graph_from_list(graph.elements, options)
        elif not quiet:
            evaluation.message(self.get_name(), "graph", expr)

    def _evaluate_atom(self, graph, options, compute):
        head = graph.head
        if head is SymbolGraph:
            return compute(graph)
        elif head is SymbolList:
            return compute(_graph_from_list(graph.elements, options))

    def __str__(self):
        return "-Graph-"

    def get_sort_key(self, pattern_sort=False) -> tuple:
        """
        Returns a particular encoded list (which should be a tuple) that is used
        in ``Sort[]`` comparisons and in the ordering that occurs
        in an M-Expression which has the ``Orderless`` property.

        See the docstring for element.get_sort_key() for more detail.
        """

        if pattern_sort:
            return super(_NetworkXBuiltin, self).get_sort_key(True)
        else:
            # Return a sort_key tuple.
            # but with a `2` instead of `1` in the 5th position,
            # and adding two extra fields: the length in the 5th position,
            # and a hash in the 6th place.
            return [
                1,
                3,
                self.class_head_name,
                tuple(),
                2,
                len(self.vertices),
                hash(self),
            ]
            return hash(self)


class _FullGraphRewrite(Exception):
    pass


def _normalize_edges(edges):
    for edge in edges:
        if edge.has_form("System`Property", 2):
            expr, prop = edge.elements
            yield Expression(edge.get_head(), list(_normalize_edges([expr]))[0], prop)
        elif edge.get_head_name() == "System`Rule":
            yield Expression(SymbolDirectedEdge, *edge.elements)
        else:
            yield edge


def is_connected(G):
    if len(G) == 0:  # empty graph?
        return True
    elif G.is_directed():
        return nx.is_strongly_connected(G)
    else:
        return nx.is_connected(G)


def _edge_weights(options):
    expr = options.get("Pymathics`EdgeWeight")
    if expr is None:
        return []
    if not expr.has_form("List", None):
        return []
    return expr.elements


class _GraphParseError(Exception):
    def __init__(self, msg=""):
        self.msg = msg

    def __repr__(self):
        if self.msg:
            return "GraphParseError: " + self.msg
        else:
            return "GraphParseError."


class Graph(Atom):
    class_head_name = "Pymathics`Graph"

    options = DEFAULT_GRAPH_OPTIONS

    def __init__(self, Gr, **kwargs):
        super(Graph, self).__init__()
        self.G = Gr
        self.mixed = kwargs.get("mixed", False)

        # Here we define types that appear on some, but not all
        # graphs. So these are optional, which we given an initial
        # value of None

        # Some graphs has a second int parameter like HknHarary
        self.n: Optional[int] = None

        # Trees have a root
        self.root: Optional[int] = None

    def __hash__(self):
        return hash(("Pymathics`Graph", self.G))

    def __str__(self):
        return "-Graph-"

    def atom_to_boxes(self, f, evaluation) -> "GraphBox":
        return GraphBox(self.G)

    def add_edges(self, new_edges, new_edge_properties):
        G = self.G.copy()
        mathics_new_edges = list(_normalize_edges(new_edges))
        return _create_graph(
            mathics_new_edges, new_edge_properties, options={}, from_graph=G
        )

    def add_vertices(self, *vertices_to_add):
        G = self.G.copy()
        G.add_nodes_from(vertices_to_add)
        return Graph(G)

    def coalesced_graph(self, evaluation):
        if not isinstance(self.G, (nx.MultiDiGraph, nx.MultiGraph)):
            return self.G, "WEIGHT"

        new_edges = defaultdict(lambda: 0)
        for u, v, w in self.G.edges.data("Pymathics`EdgeWeight", default=None):
            if w is not None:
                w = float(w.evaluate(evaluation).to_mpmath())
            else:
                w = 1
            new_edges[(u, v)] += w

        if self.G.is_directed():
            new_graph = nx.DiGraph()
        else:
            new_graph = nx.Graph()

        new_graph.add_edges_from(
            ((u, v, {"WEIGHT": w}) for (u, v), w in new_edges.items())
        )

        return new_graph, "WEIGHT"

    def delete_edges(self, edges_to_delete) -> "Graph":
        G = self.G.copy()
        directed = G.is_directed()

        normalized_edges_to_delete = list(_normalize_edges(edges_to_delete))

        for edge in normalized_edges_to_delete:
            if edge.has_form("DirectedEdge", 2):
                if directed:
                    u, v = edge.elements
                    G.remove_edge(u, v)
            elif edge.has_form("UndirectedEdge", 2):
                u, v = edge.elements
                if directed:
                    G.remove_edge(u, v)
                    G.remove_edge(v, u)
                else:
                    G.remove_edge(u, v)

        return Graph(G)

    def delete_vertices(self, vertices_to_delete):
        G = self.G.copy()
        for n in vertices_to_delete:
            G.remove_node(n)
        return Graph(G)

    def default_format(self, evaluation, form):
        return "-Graph-"

    def do_format(self, evaluation, form):
        return self

    @property
    def edges(self) -> tuple:
        # TODO: check if this should not return expressions
        return self.G.edges

    def empty(self):
        return len(self.G) == 0

    @property
    def head(self):
        return SymbolGraph

    def is_directed(self):
        if self.G.is_directed():
            return not self.mixed
        return False

    def is_loop_free(self):
        return not any(True for _ in nx.nodes_with_selfloops(self.G))

    # networkx graphs can't be used for mixed
    def is_mixed_graph(self):
        return self.mixed
        # return self.edges. ... is_mixed()

    def is_multigraph(self):
        return isinstance(self.G, (nx.MultiDiGraph, nx.MultiGraph))

    def get_sort_key(self, pattern_sort=False) -> tuple:
        """
        Returns a particular encoded list (which should be a tuple) that is used
        in ``Sort[]`` comparisons and in the ordering that occurs
        in an M-Expression which has the ``Orderless`` property.

        See the docstring for element.get_sort_key() for more detail.
        """

        if pattern_sort:
            return super(Graph, self).get_sort_key(True)
        else:
            # Return a sort_key tuple.
            # but with a `2` instead of `1` in the 5th position,
            # and adding two extra fields: the length in the 5th position,
            # and a hash in the 6th place.
            return [
                1,
                3,
                self.class_head_name,
                tuple(),
                2,
                len(self.vertices),
                hash(self),
            ]

    def sort_vertices(self, vertices):
        return sorted(vertices)

    def update_weights(self, evaluation):
        weights = None
        G = self.G

        if self.is_multigraph():
            for u, v, k, w in G.edges.data(
                "Pymathics`EdgeWeight", default=None, keys=True
            ):
                data = G.get_edge_data(u, v, key=k)
                w = data.get()
                if w is not None:
                    w = w.evaluate(evaluation).to_mpmath()
                    G[u][v][k]["WEIGHT"] = w
                    weights = "WEIGHT"
        else:
            for u, v, w in G.edges.data("Pymathics`EdgeWeight", default=None):
                if w is not None:
                    w = w.evaluate(evaluation).to_mpmath()
                    G[u][v]["WEIGHT"] = w
                    weights = "WEIGHT"

        return weights

    @property
    def value(self):
        return self.G

    @property
    def vertices(self):
        return self.G.nodes


def _parse_item(x, attr_dict=None):
    if x.has_form("Property", 2):
        expr, prop = x.elements
        attr_dict = _parse_property(prop, attr_dict)
        return _parse_item(expr, attr_dict)
    else:
        return x, attr_dict


def _graph_from_list(rules, options, new_vertices=None):
    if not rules:
        return Graph(nx.Graph())
    else:
        new_edges, new_edge_properties = zip(*[_parse_item(x) for x in rules])
        return _create_graph(
            new_edges, new_edge_properties, options=options, new_vertices=new_vertices
        )


def _create_graph(
    new_edges, new_edge_properties, options, from_graph=None, new_vertices=None
):
    vertices_dict = {}
    # Classification of vertex and edges
    known_vertices = set()
    vertices = []
    vertex_properties = []

    def add_vertex(x, attr_dict=None):
        if x.has_form("Property", 2):
            expr, prop = x.elements
            attr_dict = _parse_property(prop, attr_dict)
            return add_vertex(expr, attr_dict)
        elif x not in known_vertices:
            known_vertices.add(x)
            vertices.append(x)
            vertex_properties.append(attr_dict)
        return x

    directed_edges = []
    undirected_edges = []

    if from_graph is not None:
        old_vertices = dict(from_graph.nodes.data())
        vertices += old_vertices
        edges = list(from_graph.edges.data())

        for edge, attr_dict in edges:
            u, v = edge.elements
            if edge.head in (SymbolDirectedEdge, SymbolRule):
                directed_edges.append((u, v, attr_dict))
            else:
                undirected_edges.append((u, v, attr_dict))

        multigraph = [from_graph.is_multigraph()]
    else:
        edges = []
        edge_properties = []

        multigraph = [False]

    if new_vertices is not None:
        for v in new_vertices:
            add_vertex(v)

    def add_vertex(x, attr_dict=None):
        if attr_dict is None:
            attr_dict = {}
        if x.has_form("Pymathics`Property", 2):
            expr, prop = x.elements
            attr_dict.update(_parse_property(prop, attr_dict))
            return add_vertex(expr, attr_dict)
        elif x not in known_vertices:
            known_vertices.add(x)
            vertices.append(x)
            vertex_properties.append(attr_dict)
            vertices_dict[x] = attr_dict
        else:
            vertices_dict[x].update(attr_dict)
        return x

    if new_vertices is not None:
        for v in new_vertices:
            add_vertex(v)

    known_edges = set(edges)
    # It is simpler to just recompute this than change the above to work
    # incrementally
    known_vertices = set(vertices)

    def track_edges(*edges):
        if multigraph[0]:
            return
        previous_n_edges = len(known_edges)
        for edge in edges:
            known_edges.add(edge)
        if len(known_edges) < previous_n_edges + len(edges):
            multigraph[0] = True

    edge_weights = _edge_weights(options)
    use_directed_edges = options.get("System`DirectedEdges", SymbolTrue) is SymbolTrue

    directed_edge_head = (
        SymbolDirectedEdge if use_directed_edges else SymbolUndirectedEdge
    )

    def parse_edge(r, attr_dict=None):
        if attr_dict is None:
            attr_dict = {}

        if isinstance(r, Atom):
            raise _GraphParseError(
                msg=f"{r} is an atom, and hence does not define an edge."
            )

        if r.has_form("Pymathics`Property", None):
            expr, prop = r.elements
            attr_dict.update(_parse_property(prop, attr_dict))
            return parse_edge(expr, attr_dict)

        if r.head not in (
            SymbolRule,
            SymbolDirectedEdge,
            SymbolTwoWayRule,
            SymbolUndirectedEdge,
        ):
            raise _GraphParseError(msg=f"{r} is not an edge description.")

        r_head = r.head
        elements = r.elements

        if len(elements) != 2:
            raise _GraphParseError(
                msg=f"{r} does not have 2 elements, so it is not an edge."
            )

        u, v = elements
        assert isinstance(u, BaseElement) and isinstance(v, BaseElement)
        u = add_vertex(u)
        v = add_vertex(v)

        if r_head in (SymbolRule, SymbolDirectedEdge):
            edges_container = directed_edges
            head = directed_edge_head
            track_edges((u, v))
        elif r_head in (SymbolTwoWayRule, SymbolUndirectedEdge):
            edges_container = undirected_edges
            head = SymbolUndirectedEdge
            track_edges((u, v), (v, u))
        else:
            raise _GraphParseError(msg=f"{r_head} is an unknown kind of edge.")

        if r_head is head:
            edges.append(r)
        else:
            edges.append(Expression(head, u, v))
        edge_properties.append(attr_dict)

        edges_container.append((u, v, attr_dict))

    try:

        def full_new_edge_properties(new_edge_style):
            for i, (attr_dict, w) in enumerate(zip(new_edge_properties, edge_weights)):
                attr_dict = {} if attr_dict is None else attr_dict.copy()
                attr_dict["Pymathics`EdgeWeight"] = from_python(w)
                yield attr_dict
            # FIXME: figure out what to do here. Color is a mess.
            # for i, (attr_dict, s) in enumerate(zip(new_edge_style, new_edge_style)):
            #     attr_dict = {} if attr_dict is None else attr_dict.copy()
            #     attr_dict["Pymathics`EdgeStyle"] = s
            #     yield attr_dict
            for attr_dict in new_edge_properties[len(edge_weights) :]:
                yield attr_dict

        if "Pymathics`EdgeStyle" in options:
            # FIXME: Figure out what to do here:
            # Color is a f-ing mess.
            # edge_options = options["Pymathics`EdgeStyle"].to_python()
            edge_options = []
        else:
            edge_options = []
        edge_properties = list(full_new_edge_properties(edge_options))
        for edge, attr_dict in zip(new_edges, edge_properties):
            parse_edge(edge, attr_dict)
    except _GraphParseError:
        return None

    empty_dict = {}
    mixed = False
    if directed_edges:
        G = nx.MultiDiGraph() if multigraph[0] else nx.DiGraph()
        nodes_seen = set()
        for u, v, attr_dict in directed_edges:
            attr_dict = attr_dict or empty_dict
            G.add_edge(u, v, **attr_dict)
            nodes_seen.add(u)
            nodes_seen.add(v)

        unseen_vertices = set(vertices) - nodes_seen
        for v in unseen_vertices:
            G.add_node(v)

        if undirected_edges:
            mixed = True
        for u, v, attr_dict in undirected_edges:
            attr_dict = attr_dict or empty_dict
            G.add_edge(u, v, **attr_dict)
            G.add_edge(v, u, **attr_dict)
    else:
        G = nx.MultiGraph() if multigraph[0] else nx.Graph()
        for u, v, attr_dict in undirected_edges:
            attr_dict = attr_dict or empty_dict
            G.add_edge(u, v, **attr_dict)

    # For what is this?
    _EdgeCollection(
        edges,
        edge_properties,
        n_directed=len(directed_edges),
        n_undirected=len(undirected_edges),
    )

    g = Graph(G, mixed=mixed)
    _process_graph_options(g, options)
    return g


class _PatternList(_NetworkXBuiltin):
    def eval(self, graph, expression, evaluation, options):
        "%(name)s[graph_, OptionsPattern[%(name)s]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if graph:
            return ListExpression(*(from_python(q) for q in self._items(graph)))

    def eval_patt(self, graph, patt, expression, evaluation, options):
        "%(name)s[graph_, patt_, OptionsPattern[%(name)s]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if graph:
            return Expression(SymbolCases, ListExpression(*self._items(graph)), patt)


class AdjacencyList(_NetworkXBuiltin):
    """
    <url>
    :Adjacency list:
    https://en.wikipedia.org/wiki/Adjacency_list</url> (<url>
    :NetworkX:
    https://networkx.org/documentation/networkx-2.8.8/reference/readwrite/adjlist.html</url>, <url>
    :WMA:
    https://reference.wolfram.com/language/ref/AdjacencyList.html</url>)

    <dl>
      <dt>'AdjacencyList'[$graph$, $v$]
      <dd>gives a list of vertices adjacent to $v$ in a 'Graph' \
          or a list of edges $g$.

      <dt>'AdjacencyList'[$graph$, $pattern$]
      <dd>gives a list of vertices adjacent to vertices \
          matching $pattern$.
    </dl>

    >> AdjacencyList[{1 -> 2, 2 -> 3}, 3]
     = {2}

    >> AdjacencyList[{1 -> 2, 2 -> 3}, _?EvenQ]
     = {1, 3}

    >> AdjacencyList[{x -> 2, x -> 3, x -> 4, 2 -> 10, 2 -> 11, 4 -> 20, 4 -> 21, 10 -> 100}, 10, 2]
     = {2, 11, 100, x}
    """

    summary_text = "list the adjacent vertices"

    def _retrieve(self, graph, what, neighbors, expression, evaluation):
        if what.get_head_name() in pattern_objects:
            collected = set()
            match = Matcher(what).match
            for v in graph.G.nodes:
                if match(v, evaluation):
                    collected.update(neighbors(v))
            return ListExpression(*sorted(collected))
        elif graph.G.has_node(what):
            return ListExpression(*sorted(neighbors(what)))
        else:
            self._not_a_vertex(expression, 2, evaluation)

    def eval(self, graph, what, expression, evaluation, options):
        "%(name)s[graph_, what_, OptionsPattern[%(name)s]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if graph:
            G = graph.G.to_undirected()  # FIXME inefficient
            return self._retrieve(
                graph, what, lambda v: G.neighbors(v), expression, evaluation
            )

    def eval_d(self, graph, what, d, expression, evaluation, options):
        "%(name)s[graph_, what_, d_, OptionsPattern[%(name)s]]"
        py_d = d.to_mpmath()
        if py_d is None:
            return

        graph = self._build_graph(graph, evaluation, options, expression)
        if graph:
            G = graph.G

            def neighbors(v):
                return nx.ego_graph(
                    G, v, radius=py_d, undirected=True, center=False
                ).nodes()

            return self._retrieve(graph, what, neighbors, expression, evaluation)


class DirectedEdge(Builtin):
    """
    Edge of a <url>
    :Directed graph:
    https://en.wikipedia.org/wiki/Directed_graph</url> (<url>
    :NetworkX:
    https://networkx.org/documentation/networkx-2.8.8/reference/classes/digraph.html</url>, <url>
    :WMA:
    https://reference.wolfram.com/language/ref/DirectedEdge.html</url>)


    <dl>
      <dt>'DirectedEdge[$u$, $v$]'
      <dd>create a directed edge from $u$ to $v$.
    </dl>
    """

    summary_text = "make a directed graph edge"


class EdgeConnectivity(_NetworkXBuiltin):
    """
    <url>
    :Edge connectivity:
    https://en.wikipedia.org/wiki/Directed_graph#Directed_graph_connectivity</url> (<url>
    :NetworkX:
    https://networkx.org/documentation/networkx-2.8.8/reference/algorithms/\
generated/networkx.algorithms.connectivity.connectivity.edge_connectivity.html</url>, <url>
    :WMA:
    https://reference.wolfram.com/language/ref/EdgeConnectivity.html</url>)

    <dl>
      <dt>'EdgeConnectivity[$g$]'
      <dd>gives the edge connectivity of the graph $g$.
    </dl>

    >> EdgeConnectivity[{1 <-> 2, 2 <-> 3}]
     = 1

    >> EdgeConnectivity[{1 -> 2, 2 -> 3}]
     = 0

    >> EdgeConnectivity[{1 -> 2, 2 -> 3, 3 -> 1}]
     = 1

    >> EdgeConnectivity[{1 <-> 2, 2 <-> 3, 1 <-> 3}]
     = 2

    >> EdgeConnectivity[{1 <-> 2, 3 <-> 4}]
     = 0

    #> EdgeConnectivity[Graph[{}]]
     = EdgeConnectivity[-Graph-]
    """

    summary_text = "edge connectivity of a graph"

    def eval(self, graph, expression, evaluation, options):
        "%(name)s[graph_, OptionsPattern[%(name)s]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if graph and not graph.empty():
            return Integer(nx.edge_connectivity(graph.G))

    def eval_st(self, graph, s, t, expression, evaluation, options):
        "%(name)s[graph_, s_, t_, OptionsPattern[%(name)s]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if graph and not graph.empty():
            return Integer(nx.edge_connectivity(graph.G, s, t))


class EdgeIndex(_NetworkXBuiltin):
    """
    <url>
    :WMA link:
    https://reference.wolfram.com/language/ref/EdgeIndex.html</url>

    <dl>
    <dt>'EdgeIndex['graph', 'edge']'
    <dd>gives the position of the 'edge' in the list of edges associated \
    to the graph.
    </dl>
    """

    summary_text = "find the position of an edge"

    def eval(self, graph, v, expression, evaluation, options):
        "%(name)s[graph_, v_, OptionsPattern[%(name)s]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if graph:
            # FIXME: check if directionality must be considered or not.
            try:
                i = list(graph.edges).index(v.elements)
            except Exception:
                self._not_an_edge(expression, Integer2, evaluation)
                return
            return Integer(i + 1)


class EdgeList(_PatternList):
    """
    <url>
    :WMA link:
    https://reference.wolfram.com/language/ref/EdgeList.html</url>

    <dl>
      <dt>'EdgeList'[$g$]
      <dd>gives the list of edges that defines $g$
    </dl>
    """

    summary_text = "list the edges of a graph"

    def _items(self, graph):
        return graph.edges


class EdgeRules(_NetworkXBuiltin):
    """
    <url>
    :WMA link:
    https://reference.wolfram.com/language/ref/EdgeRules.html</url>

    <dl>
      <dt>'EdgeRules'[$g$]
      <dd> gives the list of edge rules for the graph $g$.
    </dl>
    """

    summary_text = "list the edge as rules"

    def eval(self, graph, expression, evaluation, options):
        "%(name)s[graph_, OptionsPattern[%(name)s]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if graph:

            def rules():
                for edge in graph.edges:
                    u, v = edge
                    yield Expression(SymbolRule, u, v)

            return ListExpression(*list(rules()))


class FindShortestPath(_NetworkXBuiltin):
    """

    <url>
    :Shortest path problem:
    https://en.wikipedia.org/wiki/Shortest_path_problem</url> (<url>
    :NetworkX:
    https://networkx.org/documentation/networkx-2.8.8/reference/algorithms\
    /generated/networkx.algorithms.shortest_paths.generic.shortest_path.html</url>, <url>
    :WMA:
    https://reference.wolfram.com/language/ref/FindShortestPath.html</url>)

    <dl>
      <dt>'FindShortestPath'[$g$, $src$, $tgt$]
      <dd>List the vertices in the shortest path connecting the source $src$ with the target $tgt$ \
          in the graph $g$.
    </dl>


    >> FindShortestPath[{1 <-> 2, 2 <-> 3, 3 <-> 4, 2 <-> 4, 4 -> 5}, 1, 5]
     = {1, 2, 4, 5}

    >> FindShortestPath[{1 <-> 2, 2 <-> 3, 3 <-> 4, 4 -> 2, 4 -> 5}, 1, 5]
     = {1, 2, 3, 4, 5}

    >> FindShortestPath[{1 <-> 2, 2 <-> 3, 4 -> 3, 4 -> 2, 4 -> 5}, 1, 5]
     = {}

    >> g = Graph[{1 -> 2, 2 -> 3, 1 -> 3}, EdgeWeight -> {0.5, a, 3}];

    #> FindShortestPath[{}, 1, 2]
     : The vertex at position 2 in FindShortestPath[{}, 1, 2] does not belong to the graph at position 1.
     = FindShortestPath[{}, 1, 2]

    #> FindShortestPath[{1 -> 2}, 1, 3]
     : The vertex at position 3 in FindShortestPath[{1 -> 2}, 1, 3] does not belong to the graph at position 1.
     = FindShortestPath[{1 -> 2}, 1, 3]
    """

    summary_text = "find the shortest path between two vertices"

    def eval_s_t(self, graph, s, t, expression, evaluation, options):
        "FindShortestPath[graph_, s_, t_, OptionsPattern[FindShortestPath]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if not graph:
            return
        G = graph.G
        if not G.has_node(s):
            self._not_a_vertex(expression, 2, evaluation)
        elif not G.has_node(t):
            self._not_a_vertex(expression, 3, evaluation)
        else:
            try:
                weight = graph.update_weights(evaluation)
                return ListExpression(
                    *list(nx.shortest_path(G, source=s, target=t, weight=weight)),
                )
            except nx.exception.NetworkXNoPath:
                return ListExpression()


class FindVertexCut(_NetworkXBuiltin):
    """
    <url>:Minimum cut:https://en.wikipedia.org/wiki/Minimum_cut</url>\
    (<url>
    :NetworkX:
    https://networkx.org/documentation/networkx-2.8.8/reference/algorithms/\
generated/networkx.algorithms.connectivity.cuts.minimum_node_cut.html
    </url>, \
    <url>
    :WMA:
    https://reference.wolfram.com/language/ref/FindVertexCut.html</url>)

    <dl>
    <dt>'FindVertexCut[$g$]'
        <dd>finds a set of vertices of minimum cardinality that, if removed, renders $g$ disconnected.
    <dt>'FindVertexCut[$g$, $s$, $t$]'
        <dd>finds a vertex cut that disconnects all paths from $s$ to $t$.
    </dl>

    >> g = Graph[{1 -> 2, 2 -> 3}]; FindVertexCut[g]
     = {}

    >> g = Graph[{1 <-> 2, 2 <-> 3}]; FindVertexCut[g]
     = {2}

    >> g = Graph[{1 <-> x, x <-> 2, 1 <-> y, y <-> 2, x <-> y}]; FindVertexCut[g]
     = {x, y}

    #> FindVertexCut[Graph[{}]]
     = {}
    #> FindVertexCut[Graph[{}], 1, 2]
     : The vertex at position 2 in FindVertexCut[-Graph-, 1, 2] does not belong to the graph at position 1.
     = FindVertexCut[-Graph-, 1, 2]
    """

    summary_text = "find the vertex cuts"

    def eval(self, graph, expression, evaluation, options):
        "FindVertexCut[graph_, OptionsPattern[%(name)s]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if graph:
            if graph.empty() or not is_connected(graph.G):
                return ListExpression()
            else:
                return ListExpression(
                    *graph.sort_vertices(nx.minimum_node_cut(graph.G))
                )

    def eval_st(self, graph, s, t, expression, evaluation, options):
        "FindVertexCut[graph_, s_, t_, OptionsPattern[%(name)s]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if not graph:
            return

        G = graph.G
        if not G.has_node(s):
            self._not_a_vertex(expression, 2, evaluation)
        elif not G.has_node(t):
            self._not_a_vertex(expression, 3, evaluation)
        elif graph.empty() or not is_connected(graph.G):
            return ListExpression()
        else:
            return ListExpression(*graph.sort_vertices(nx.minimum_node_cut(G, s, t)))


class GraphAtom(AtomBuiltin):
    """
    <url>:Graph:https://en.wikipedia.org/wiki/graph</url> (<url>:WMA:
    https://reference.wolfram.com/language/ref/Graph.html</url>)
    <dl>
      <dt>'Graph[{$e1, $e2, ...}]'
      <dd>returns a graph with edges $e_j$.
    </dl>

    <dl>
      <dt>'Graph[{v1, v2, ...}, {$e1, $e2, ...}]'
      <dd>returns a graph with vertices $v_i$ and edges $e_j$.
    </dl>

    >> Graph[{1->2, 2->3, 3->1}]
     = -Graph-

    #>> Graph[{1->2, 2->3, 3->1}, EdgeStyle -> {Red, Blue, Green}]
    # = -Graph-

    >> Graph[{1->2, Property[2->3, EdgeStyle -> Thick], 3->1}]
     = -Graph-

    # >> Graph[{1->2, 2->3, 3->1}, VertexStyle -> {1 -> Green, 3 -> Blue}]
    # = -Graph-

    >> Graph[x]
     = Graph[x]

    >> Graph[{1}]
     = Graph[{1}]

    >> Graph[{{1 -> 2}}]
     = Graph[{{1 -> 2}}]

    """

    # requires = ("networkx",)
    name = "Graph"
    options = DEFAULT_GRAPH_OPTIONS
    summary_text = "build a graph"

    def eval(self, graph, evaluation, options):
        "Pymathics`Graph[graph_List, OptionsPattern[%(name)s]]"
        return _graph_from_list(graph.elements, options)

    def eval_1(self, vertices, edges, evaluation, options):
        "Pymathics`Graph[vertices_List, edges_List, OptionsPattern[%(name)s]]"
        return _graph_from_list(
            edges.elements, options=options, new_vertices=vertices.elements
        )


class HighlightGraph(_NetworkXBuiltin):
    """
    <url>
    :WMA link:
    https://reference.wolfram.com/language/ref/HighlightGraph.html</url>

    <dl>
    <dt>'HighlightGraph'[$graph$, $what$]
    <dd>highlight in $graph$ the elements enumerated in $what$ by adding style marks.
    </dl>
    """

    summary_text = "highlight elements in a graph"

    def eval(self, graph, what, expression, evaluation, options):
        "HighlightGraph[graph_, what_List, OptionsPattern[%(name)s]]"
        default_highlight = [Expression(SymbolRGBColor, Integer1, Integer0, Integer0)]

        def parse(item):
            if item.get_head_name() == "System`Rule":
                return Expression(SymbolDirectedEdge, *item.elements)
            else:
                return item

        rules = []
        for element in what.elements:
            if element.get_head_name() == "System`Style":
                if len(element.elements) >= 2:
                    rules.append((parse(element.elements[0]), element.elements[1:]))
            else:
                rules.append((parse(element), default_highlight))

        rules.append((Expression(SymbolBlank), Expression(SymbolMissing)))

        graph = self._build_graph(graph, evaluation, options, expression)
        if graph:
            rule_exprs = ListExpression(*[Expression(SymbolRule, *r) for r in rules])
            return graph.with_highlight(rule_exprs)


class Property(Builtin):
    """
    <url>
    :WMA link:
    https://reference.wolfram.com/language/ref/Property.html</url>

    <dl>
      <dt>'Property'[$item$, {$name$, $val$}]
      <dd>associate a property called $name$ with value $val$ to $item$.
    </dl>

    """

    summary_text = "assign a property"


class PropertyValue(Builtin):
    """
    <url>
    :WMA link:
    https://reference.wolfram.com/language/ref/PropertyValue.html</url>

    <dl>
      <dt>'PropertyValue'[{$obj$, $item$}, $name$]
      <dd>gives the value of a property associated with the name  $name$
          for $item$ in the object $obj$.
    </dl>


    >> g = Graph[{a <-> b, Property[b <-> c, SomeKey -> 123]}];
    >> PropertyValue[{g, b <-> c}, SomeKey]
     = 123
    >> PropertyValue[{g, b <-> c}, SomeUnknownKey]
     = $Failed
    """

    requires = ("networkx",)
    summary_text = "retrieve the value of a property"

    def eval(self, graph, item, name, evaluation):
        "PropertyValue[{graph_Graph, item_}, name_Symbol]"
        name_str = name.get_name()
        if isinstance(graph, Graph):
            if (
                item.has_form("Rule", 2)
                or item.has_form("DirectedEdge", 2)
                or item.has_form("UndirectedEdge", 2)
            ):
                item_g = graph.G.edges().get(item.elements)
            else:
                item_g = graph.G.nodes().get(item)

            if item_g is None:
                return SymbolFailed

            value = item_g.get(name_str, SymbolFailed)
            return value


class VertexAdd(_NetworkXBuiltin):
    """
    <url>
    :WMA link:
    https://reference.wolfram.com/language/ref/VertexAdd.html</url>

    <dl>
      <dt>'VertexAdd'[$g$, $ver$]
      <dd>create a new graph from $g$, by adding the vertex $ver$.
    </dl>

    >> g1 = Graph[{1 -> 2, 2 -> 3}];
    >> g2 = VertexAdd[g1, 4]
     = -Graph-
    >> g3 = VertexAdd[g2, {5, 10}]
     = -Graph-
    >> VertexAdd[{a -> b}, c]
     = -Graph-
    """

    summary_text = "add a vertex"

    def eval(self, graph: Expression, what, expression, evaluation, options):
        "VertexAdd[graph_, what_, OptionsPattern[VertexAdd]]"
        mathics_graph = self._build_graph(graph, evaluation, options, expression)
        if mathics_graph:
            if what.get_head_name() == "System`List":
                return mathics_graph.add_vertices(
                    *zip(*[_parse_item(x) for x in what.elements])
                )
            else:
                return mathics_graph.add_vertices(*zip(*[_parse_item(what)]))


class VertexConnectivity(_NetworkXBuiltin):
    """
    <url>
    :WMA link:
    https://reference.wolfram.com/language/ref/VertexConnectivity.html</url>

    <dl>
      <dt>'VertexConnectivity'[$g$]
      <dd>gives the vertex connectivity of the graph $g$.
    </dl>

    >> VertexConnectivity[{1 <-> 2, 2 <-> 3}]
     = 1

    >> VertexConnectivity[{1 -> 2, 2 -> 3}]
     = 0

    >> VertexConnectivity[{1 -> 2, 2 -> 3, 3 -> 1}]
     = 1

    >> VertexConnectivity[{1 <-> 2, 2 <-> 3, 1 <-> 3}]
     = 2

    >> VertexConnectivity[{1 <-> 2, 3 <-> 4}]
     = 0

    #> VertexConnectivity[Graph[{}]]
     = VertexConnectivity[-Graph-]
    """

    summary_text = "vertex connectivity"

    def eval(self, graph, expression, evaluation, options):
        "%(name)s[graph_, OptionsPattern[%(name)s]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if graph and not graph.empty():
            if not is_connected(graph.G):
                return Integer(0)
            else:
                return Integer(nx.node_connectivity(graph.G))

    def eval_st(self, graph, s, t, expression, evaluation, options):
        "%(name)s[graph_, s_, t_, OptionsPattern[%(name)s]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if graph and not graph.empty():
            if not is_connected(graph.G):
                return Integer(0)
            else:
                return Integer(nx.node_connectivity(graph.G, s, t))


class VertexDelete(_NetworkXBuiltin):
    """
    <url>
    :WMA link:
    https://reference.wolfram.com/language/ref/VertexDelete.html</url>

    <dl>
      <dt>'VertexDelete'[$g$, $vert$]
      <dd>remove the vertex $vert$ and their associated edges.
    </dl>

    >> g1 = Graph[{1 -> 2, 2 -> 3, 3 -> 4}];
    >> VertexDelete[g1, 3]
     = -Graph-
    >> VertexDelete[{a -> b, b -> c, c -> d, d -> a}, {a, c}]
     = -Graph-
    >> VertexDelete[{1 -> 2, 2 -> 3, 3 -> 4, 4 -> 6, 6 -> 8, 8 -> 2}, _?OddQ]
     = -Graph-
    """

    summary_text = "remove a vertex"

    def eval(self, graph, what, expression, evaluation, options) -> Optional[Graph]:
        "VertexDelete[graph_, what_, OptionsPattern[VertexDelete]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if graph:
            head_name = what.get_head_name()
            if head_name in pattern_objects:
                cases = Expression(
                    SymbolCases, ListExpression(*graph.vertices), what
                ).evaluate(evaluation)
                if cases.get_head_name() == "System`List":
                    return graph.delete_vertices(cases.elements)
            elif head_name == "System`List":
                return graph.delete_vertices(what.elements)
            else:
                return graph.delete_vertices([what])


class VertexIndex(_NetworkXBuiltin):
    """
    <url>
    :WMA link:
    https://reference.wolfram.com/language/ref/VertexIndex.html</url>
    <dl>
      <dt>'VertexIndex'['g', 'v']
      <dd> gives the integer index of the vertex 'v' in the\
       graph 'g'.
    </dl>

    >> a=.;
    >> VertexIndex[{c <-> d, d <-> a}, a]
     = 3
    """

    summary_text = "find the position of a vertex"

    def eval(self, graph, v, expression, evaluation, options):
        "%(name)s[graph_, v_, OptionsPattern[%(name)s]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if graph:
            try:
                return Integer(list(graph.vertices).index(v) + 1)
            except ValueError:
                self._not_a_vertex(expression, 2, evaluation)
        return None


class VertexList(_PatternList):
    """
    <url>
    :WMA link:
    https://reference.wolfram.com/language/ref/VertexList.html</url>
    <dl>
      <dt>'VertexList[$edgelist$]'
      <dd>list the vertices from a list of directed edges.
    </dl>

    >> a=.;
    >> VertexList[{1 -> 2, 2 -> 3}]
     = {1, 2, 3}

    >> VertexList[{a -> c, c -> b}]
     = {a, c, b}

    >> VertexList[{a -> c, 5 -> b}, _Integer -> 10]
     = {10}
    """

    summary_text = "list the vertices of a graph"

    def _items(self, graph):
        return graph.vertices


class UndirectedEdge(Builtin):
    """
    <url>
    :WMA link:
    https://reference.wolfram.com/language/ref/UndirectedEdge.html</url>

    <dl>
      <dt>'UndirectedEdge[$u$, $v$]'
      <dd>create an undirected edge between $u$ and $v$.
    </dl>

    >> a <-> b
     = UndirectedEdge[a, b]

    >> (a <-> b) <-> c
     = UndirectedEdge[UndirectedEdge[a, b], c]

    >> a <-> (b <-> c)
     = UndirectedEdge[a, UndirectedEdge[b, c]]
    """

    summary_text = "undirected graph edge"
    pass


# class EdgeAdd(_NetworkXBuiltin):
#     """
#     >> EdgeAdd[{1->2,2->3},3->1]
#      = -Graph-
#     """

#     def eval(self, graph: Expression, what, expression, evaluation, options):
#         "%(name)s[graph_, what_, OptionsPattern[%(name)s]]"
#         mathics_graph = self._build_graph(graph, evaluation, options, expression)
#         if mathics_graph:
#             if what.get_head_name() == "System`List":
#                 return mathics_graph.add_edges(*zip(*[_parse_item(x) for x in what.elements]))
#             else:
#                 return mathics_graph.add_edges(*zip(*[_parse_item(what)]))


class EdgeDelete(_NetworkXBuiltin):
    """
    Delete an Edge (<url>
    :WMA:
    https://reference.wolfram.com/language/ref/EdgeDelete.html</url>)

    <dl>
      <dt>'EdgeDelete'[$g$, $edge$]
      <dd>remove the edge $edge$.
    </dl>

    >> g = Graph[{1 -> 2, 2 -> 3, 3 -> 1}, VertexLabels->True]
     = -Graph-

    >> EdgeList[EdgeDelete[g, 2 -> 3]]
     = {{1, 2}, {3, 1}}

    ## >> g = Graph[{4<->5,5<->7,7<->9,9<->5,2->4,4->6,6->2}, VertexLabels->True]
    ##  = -Graph-

    ## >> EdgeDelete[g, _UndirectedEdge]
    ##  = -Graph-
    """

    summary_text = "remove an edge"

    def eval(self, graph, what, expression, evaluation, options) -> Optional[Graph]:
        "EdgeDelete[graph_, what_, OptionsPattern[EdgeDelete]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if graph:
            head_name = what.get_head_name()
            if head_name in pattern_objects:
                cases = Expression(
                    SymbolCases, to_mathics_list(*graph.edges), what
                ).evaluate(evaluation)
                if cases.get_head_name() == "System`List":
                    return graph.delete_edges(cases.elements)
            elif head_name == "System`List":
                return graph.delete_edges(what.elements)
            else:
                return graph.delete_edges([what])



# -*- coding: utf-8 -*-
"""
Unit tests for pymathics.graph.components
"""
from test.helper import check_evaluation, evaluate, evaluate_value


def setup_module(module):
    """Load pymathics.graph"""
    assert evaluate_value('LoadModule["pymathics.graph"]') == "pymathics.graph"
    evaluate("SortList[list_] := Sort[Map[Sort, list]]")


def test_connected_components():
    for str_expr, str_expected in [
        ("PlanarGraphQ[Graph[{}]]", "False"),
        ("g = Graph[{1 -> 2, 2 -> 3, 3 <-> 4}];", "Null"),
        ("SortList[ConnectedComponents[g]]", "{{1}, {2}, {3, 4}}"),
        ("g = Graph[{1 -> 2, 2 -> 3, 3 -> 1}];", "Null"),
        ("SortList[ConnectedComponents[g]]", "{{1, 2, 3}}"),
    ]:
        check_evaluation(str_expr, str_expected)



"""
Unit tests for pymathics.graph.measures_and_metrics
"""

from test.helper import check_evaluation, evaluate_value


def setup_module(module):
    """Load pymathics.graph"""
    assert evaluate_value('LoadModule["pymathics.graph"]') == "pymathics.graph"


def test_AcyclicQ():
    for str_expr, str_expected, mess in [
        (
            "AcyclicGraphQ[Graph[{1 -> 2, 2 -> 3, 5 -> 2, 3 -> 4, 5 -> 3}]]",
            "True",
            None,
        ),
        (
            "AcyclicGraphQ[Graph[{1 -> 2, 2 -> 3, 5 -> 2, 3 -> 4, 5 <-> 3}]]",
            "False",
            None,
        ),
        (
            "AcyclicGraphQ[Graph[{1 <-> 2, 2 <-> 3, 5 <-> 2, 3 <-> 4, 5 <-> 3}]]",
            "False",
            None,
        ),
        # (
        #     "AcyclicGraphQ[Graph[{}]]",
        #     "False",
        #     None,
        # ),
        (
            'AcyclicGraphQ["abc"]',
            "False",
            ["Expected a graph at position 1 in AcyclicGraphQ[abc]."],
        ),
    ]:
        check_evaluation(str_expr, str_expected, expected_messages=mess)



"""
Graphs - Vertices and Edges


A Graph is a tuple of a set of Nodes and Edges.

Mathics3 Module that provides functions and variables for working with graphs.

Examples:

   >> LoadModule["pymathics.graph"]
    = pymathics.graph

   >> BinomialTree[3, DirectedEdges->True]
     = -Graph-

   >> BalancedTree[3, 3]
     = -Graph-

   >> g = Graph[{1 -> 2, 2 -> 3, 3 <-> 4}, VertexLabels->True]
    = -Graph-

   >> ConnectedComponents[g]
    = {{3, 4}, {2}, {1}}

   >> WeaklyConnectedComponents[g]
    = {{1, 2, 3, 4}}

   >> GraphDistance[g, 1, 4]
    = 3

   >> GraphDistance[g, 3, 2]
    = Infinity

<url>
:NetworkX:
https://networkx.org</url> does the heavy lifting here.
"""

from pymathics.graph.base import (
    AdjacencyList,
    DirectedEdge,
    EdgeConnectivity,
    EdgeDelete,
    EdgeIndex,
    EdgeList,
    EdgeRules,
    FindShortestPath,
    FindVertexCut,
    Graph,
    GraphAtom,
    GraphBox,
    HighlightGraph,
    Property,
    PropertyValue,
    UndirectedEdge,
    VertexAdd,
    VertexConnectivity,
    VertexDelete,
    VertexIndex,
    VertexList,
)

from pymathics.graph.centralities import (
    BetweennessCentrality,
    ClosenessCentrality,
    DegreeCentrality,
    EigenvectorCentrality,
    HITSCentrality,
    KatzCentrality,
    PageRankCentrality,
)

from pymathics.graph.components import ConnectedComponents, WeaklyConnectedComponents
from pymathics.graph.curated import GraphData
from pymathics.graph.measures_and_metrics import (
    EdgeCount,
    GraphDistance,
    VertexCount,
    VertexDegree,
)
from pymathics.graph.operations import FindSpanningTree
from pymathics.graph.parametric import (
    BalancedTree,
    BarbellGraph,
    BinomialTree,
    CompleteGraph,
    CompleteKaryTree,
    CycleGraph,
    GraphAtlas,
    HknHararyGraph,
    HmnHararyGraph,
    KaryTree,
    LadderGraph,
    RandomTree,
    StarGraph,
)
from pymathics.graph.properties import (
    AcyclicGraphQ,
    ConnectedGraphQ,
    DirectedGraphQ,
    LoopFreeGraphQ,
    MixedGraphQ,
    MultigraphQ,
    PathGraphQ,
    PlanarGraphQ,
    SimpleGraphQ,
)
from pymathics.graph.random import RandomGraph
from pymathics.graph.structured import PathGraph, TreeGraph
from pymathics.graph.tree import TreeGraphAtom, TreeGraphQ
from pymathics.graph.version import __version__

pymathics_version_data = {
    "author": "The Mathics3 Team",
    "version": __version__,
    "name": "Graph",
    "requires": ["networkx"],
}

# Thsee are the publicly exported names
__all__ = [
    "AcyclicGraphQ",
    "AdjacencyList",
    "BalancedTree",
    "BarbellGraph",
    "BetweennessCentrality",
    "BinomialTree",
    "ClosenessCentrality",
    "CompleteGraph",
    "CompleteKaryTree",
    "ConnectedComponents",
    "ConnectedGraphQ",
    "CycleGraph",
    "DegreeCentrality",
    "DirectedEdge",
    "DirectedGraphQ",
    "EdgeConnectivity",
    "EdgeCount",
    "EdgeDelete",
    "EdgeIndex",
    "EdgeList",
    "EdgeRules",
    "EigenvectorCentrality",
    "FindShortestPath",
    "FindSpanningTree",
    "FindVertexCut",
    "Graph",
    "GraphAtlas",
    "GraphAtom",
    "GraphBox",
    "GraphData",
    "GraphDistance",
    "HITSCentrality",
    "HighlightGraph",
    "HknHararyGraph",
    "HmnHararyGraph",
    "KaryTree",
    "KatzCentrality",
    "LadderGraph",
    "LoopFreeGraphQ",
    "MixedGraphQ",
    "MultigraphQ",
    "PageRankCentrality",
    "PathGraph",
    "PathGraphQ",
    "PlanarGraphQ",
    "Property",
    "PropertyValue",
    "RandomGraph",
    "RandomTree",
    "SimpleGraphQ",
    "StarGraph",
    "TreeGraph",
    "TreeGraphAtom",
    "TreeGraphQ",
    "UndirectedEdge",
    "VertexAdd",
    "VertexConnectivity",
    "VertexCount",
    "VertexDegree",
    "VertexDelete",
    "VertexIndex",
    "VertexList",
    "WeaklyConnectedComponents",
    "__version__",
    "pymathics_version_data",
]



# -*- coding: utf-8 -*-

"""
GraphBox.
"""

# uses networkx

import base64
import tempfile


from typing import Tuple

from mathics.core.element import BaseElement, BoxElementMixin


from pymathics.graph.format import png_format_graph, svg_format_graph

no_doc = True

class GraphBox(BoxElementMixin):
    def __init__(self, G, **options):
        self.G = G
        self.options = options

    def boxes_to_b64text(
        self, elements: Tuple[BaseElement] = None, **options
    ) -> Tuple[bytes, Tuple[int, int]]:
        """
        Produces a base64 png representation and a tuple with the size of the pillow image
        associated to the object.
        """
        contents, size = self.boxes_to_png(elements, **options)
        encoded = base64.b64encode(contents)
        encoded = b"data:image/png;base64," + encoded
        return encoded, size

    def boxes_to_png(self, elements=None, **options) -> Tuple[bytes, Tuple[int, int]]:
        """
        returns a tuple with the set of bytes with a png representation of the image
        and the scaled size.
        """
        return png_format_graph(self.G, **self.options), (800, 600)

    def boxes_to_svg(self, elements=None, **options):
        return svg_format_graph(self.G, **self.options), (400, 300)

    def boxes_to_tex(self, elements=None, **options) -> str:
        """
        Store the associated image as a png file and return
        a LaTeX command for including it.
        """

        data, size = self.boxes_to_png(elements, **options)
        res = 100  # pixels/cm
        width_str, height_str = (str(n / res).strip() for n in size)
        head = rf"\includegraphics[width={width_str}cm,height={height_str}cm]"

        # This produces a random name, where the png file is going to be stored.
        # LaTeX does not have a native way to store an figure embeded in
        # the source.
        fp = tempfile.NamedTemporaryFile(delete=True, suffix=".png")
        path = fp.name
        fp.close()

        with open(path, "wb") as imgfile:
            imgfile.write(data)

        return head + "{" + format(path) + "}"

    def boxes_to_text(self, elements=None, **options):
        return "-Graph-"

    def boxes_to_mathml(self, elements=None, **options):
        encoded, size = self.boxes_to_b64text(elements, **options)
        decoded = encoded.decode("utf8")
        # see https://tools.ietf.org/html/rfc2397
        return f'<mglyph src="{decoded}" width="50%" height="50%" />'



# -*- coding: utf-8 -*-
"""
Graph Components and Connectivity
"""

from typing import Optional

import networkx as nx
from mathics.core.convert.expression import to_mathics_list
from mathics.core.evaluation import Evaluation
from mathics.core.list import ListExpression

from pymathics.graph.base import _NetworkXBuiltin


class ConnectedComponents(_NetworkXBuiltin):
    """
    <url>
    :Strongly connected components:
    https://en.wikipedia.org/wiki/Strongly_connected_component</url> (<url>
    :NetworkX:
    https://networkx.org/documentation/networkx-2.8.8/reference/algorithms/\
generated/networkx.algorithms.components.strongly_connected_components.html</url>, <url>
:WMA:https://reference.wolfram.com/language/ref/ConnectedComponents.html</url>)

    <dl>
      <dt>'ConnectedComponents'[$g$]
      <dd> gives the connected components of the graph $g$.
    </dl>

    >> g = Graph[{1 -> 2, 2 -> 3, 3 <-> 4}, VertexLabels->True]
     = -Graph-

    >> ConnectedComponents[g]
     = ...

    >> g = Graph[{1 -> 2, 2 -> 3, 3 -> 1}, VertexLabels->True]
     = -Graph-

    >> ConnectedComponents[g]
     = ...

    >> g = Graph[{1 <-> 2, 2 <-> 3, 3 -> 4, 4 <-> 5}, VertexLabels->True]
     = -Graph-

    >> ConnectedComponents[g]
     = ...
    """

    summary_text = "list the connected components"

    def eval(
        self, graph, expression, evaluation: Evaluation, options: dict
    ) -> Optional[ListExpression]:
        "ConnectedComponents[graph_, OptionsPattern[%(name)s]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if graph:
            connect_fn = (
                nx.strongly_connected_components
                if graph.G.is_directed()
                else nx.connected_components
            )
            components = [to_mathics_list(*sorted(c)) for c in connect_fn(graph.G)]
            return ListExpression(*components)


# This goes in Path, Cycles, Flows
# class FindHamiltonianPath(_NetworkXBuiltin):
#     """
#     <dl>
#       <dt>'FindHamiltonianPath[$g$]'
#       <dd>returns a Hamiltonian path in the given tournament graph.
#       </dl>
#
#     """
#     def eval_(self, graph, expression, evaluation: Evaluation, options):
#         "FindHamiltonianPath[graph_, OptionsPattern[FindHamiltonPath]]"

#         graph = self._build_graph(graph, evaluation: Evaluation, options, expression)
#         if graph:
#             # FIXME: for this to work we need to fill in all O(n^2) edges as an adjacency matrix?
#             path = nx.algorithms.tournament.hamiltonian_path(graph.G)
#             if path:
#                 # int_path = map(Integer, path)
#                 return to_mathics_list(*path)


class WeaklyConnectedComponents(_NetworkXBuiltin):
    """
    <url>
    :Weak components:
    https://en.wikipedia.org/wiki/Weak_component</url> (<url>
    :NetworkX:
    https://networkx.org/documentation/networkx-2.8.8/reference/algorithms/\
generated/networkx.algorithms.components.weakly_connected_components.html</url>, <url>
    :WMA:
    https://reference.wolfram.com/language/ref/WeaklyConnectedComponents.html</url>)

    <dl>
      <dt>'WeaklyConnectedComponents'[$g$]
      <dd> gives the weakly connected components of the graph $g$.
    </dl>

    >> g = Graph[{1 -> 2, 2 -> 3, 3 <-> 4}, VertexLabels->True]
     = -Graph-

    >> WeaklyConnectedComponents[g]
     = {{1, 2, 3, 4}}

    >> g = Graph[{1 -> 2, 2 -> 3, 3 -> 1}, VertexLabels->True]
     = -Graph-

    >> WeaklyConnectedComponents[g]
     = {{1, 2, 3}}

    >> g = Graph[{1 <-> 2, 2 <-> 3, 3 -> 4, 4 <-> 5, 6 <-> 7, 7 <-> 8}, VertexLabels->True]
     = -Graph-

    >> WeaklyConnectedComponents[g]
     = {{1, 2, 3, 4, 5}, {6, 7, 8}}
    """

    summary_text = "list the weakly connected components"

    def eval(self, graph, expression, evaluation: Evaluation, options):
        "WeaklyConnectedComponents[graph_, OptionsPattern[WeaklyConnectedComponents]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if graph:
            components = nx.connected_components(graph.G.to_undirected())
            result = []
            for component in components:
                result.append(sorted(component))
            return to_mathics_list(*result)



"""
Trees
"""
import networkx as nx
from mathics.core.atoms import Atom
from mathics.core.evaluation import Evaluation
from mathics.core.symbols import SymbolConstant

from pymathics.graph.base import (
    DEFAULT_GRAPH_OPTIONS,
    _graph_from_list,
    _NetworkXBuiltin,
)
from pymathics.graph.eval.tree import eval_TreeGraphQ

DEFAULT_TREE_OPTIONS = {
    **DEFAULT_GRAPH_OPTIONS,
    **{"GraphLayout": '"tree"'},
}

from mathics.core.builtin import AtomBuiltin


# FIXME: do we need to have TreeGraphAtom and TreeGraph?
# Can't these be combined into one?
class TreeGraphAtom(AtomBuiltin):
    """
    <url>:Tree Graph:https://en.wikipedia.org/wiki/Tree_(graph_theory)</url>
    (<url>:WMA:https://reference.wolfram.com/language/ref/TreeGraph.html</url>)
    <dl>
      <dt>'TreeGraph'[$edges$]
      <dd>Build a Tree-like graph from the list of edges $edges$.
      <dt>'TreeGraph'[$vert$, $edges$]
      <dd>build a Tree-like graph from the list of vertices $vert$ and  edges $edges$.
    </dl>


    >> TreeGraph[{1->2, 2->3, 2->4}]
     = -Graph-

    If the $edges$ does not match with a tree-like pattern, the evaluation fails:
    >> TreeGraph[{1->2, 2->3, 3->1}]
     : Graph is not a tree.
     = TreeGraph[{1 -> 2, 2 -> 3, 3 -> 1}]
    """

    options = DEFAULT_TREE_OPTIONS

    messages = {
        "v": "Expected first parameter vertices to be a list of vertices",
        "notree": "Graph is not a tree.",
    }
    summary_text = "build a tree graph"

    def eval(self, rules, evaluation: Evaluation, options: dict):
        "TreeGraph[rules_List, OptionsPattern[%(name)s]]"
        g = _graph_from_list(rules.elements, options)
        if not nx.is_tree(g.G):
            evaluation.message(self.get_name(), "notree")
            return

        g.G.graph_layout = "tree"
        # Compute/check/set for root?
        return g

    def eval_with_v_e(self, vertices, edges, evaluation: Evaluation, options: dict):
        "TreeGraph[vertices_List, edges_List, OptionsPattern[%(name)s]]"
        if not all(isinstance(v, Atom) for v in vertices.elements):
            evaluation.message(self.get_name(), "v")
            return

        g = _graph_from_list(
            edges.elements, options=options, new_vertices=vertices.elements
        )
        if not nx.is_tree(g.G):
            evaluation.message(self.get_name(), "notree")
            return

        g.G.graph_layout = "tree"
        # Compute/check/set for root?
        return g


class TreeGraphQ(_NetworkXBuiltin):
    """
    <url>:Tree Graph:https://en.wikipedia.org/wiki/Tree_(graph_theory)</url>
    (<url>:WMA:https://reference.wolfram.com/language/ref/TreeGraphQ.html</url>)

    <dl>
      <dt>'TreeGraphQ[$g$]'
      <dd>returns $True$ if the graph $g$ is a tree and $False$ otherwise.
    </dl>

    >> TreeGraphQ[StarGraph[3]]
     = True
    >> TreeGraphQ[CompleteGraph[2]]
     = True
    >> TreeGraphQ[CompleteGraph[3]]
     = False
    """

    summary_text = "test for a tree-like graph"

    def eval(
        self, g, expression, evaluation: Evaluation, options: dict
    ) -> SymbolConstant:
        "TreeGraphQ[g_, OptionsPattern[%(name)s]]"
        return eval_TreeGraphQ(g)



"""
Graph Properties and Measurements
"""

import networkx as nx
from mathics.core.convert.python import from_python
from mathics.core.evaluation import Evaluation
from mathics.core.symbols import SymbolFalse, SymbolTrue

from pymathics.graph.base import (
    DEFAULT_GRAPH_OPTIONS,
    Graph,
    _NetworkXBuiltin,
    is_connected,
)


class AcyclicGraphQ(_NetworkXBuiltin):
    """
    <url>
    :Acyclic graph:
    https://en.wikipedia.org/wiki/Acyclic_graph</url> test (<url>
    :NetworkX:
    https://networkx.org/documentation/stable/reference/algorithms\
/generated/networkx.algorithms.cycles.find_cycle.html
    </url>, <url>
    :WMA:
    https://reference.wolfram.com/language/ref/AcyclicGraphQ.html</url>)

    <dl>
      <dt>'AcyclicGraphQ'[$graph$]
      <dd>check if $graph$ is an acyclic graph.
    </dl>


    Create a directed graph with a cycle in it:

    >> g = Graph[{1 -> 2, 2 -> 3, 5 -> 2, 3 -> 4, 3 -> 5}, VertexLabels->True]
     = -Graph-

    >> AcyclicGraphQ[g]
     = False

    Remove a cycle edge:

    >> g = EdgeDelete[g, 5 -> 2]; EdgeList[g]
     = {{1, 2}, {2, 3}, {3, 4}, {3, 5}}

    >> AcyclicGraphQ[g]
     = True
    """

    summary_text = "test if is a graph is acyclic"

    def eval(self, graph, expression, evaluation, options):
        "AcyclicGraphQ[graph_, OptionsPattern[AcyclicGraphQ]]"
        graph = self._build_graph(graph, evaluation, options, expression, quiet=False)
        if not graph or graph.empty():
            return SymbolFalse

        try:
            cycles = nx.find_cycle(graph.G)
        except nx.exception.NetworkXNoCycle:
            return SymbolTrue
        return from_python(not cycles)


class ConnectedGraphQ(_NetworkXBuiltin):
    """
    <url>
    :Connected graph:
    https://en.wikipedia.org/wiki/Connectivity_(graph_theory)\
#Connected_vertices_and_graphs
    </url> test (<url>
    :NetworkX:
https://networkx.org/documentation/networkx-2.8.8/reference/algorithms\
/generated/networkx.algorithms.components.is_connected.html
    </url>, <url>
    :WMA:
    https://reference.wolfram.com/language/ref/ConnectedGraphQ.html</url>)

    <dl>
      <dt>'ConnectedGraphQ'[$graph$]
      <dd>check if $graph$ is a connected graph.
    </dl>

    >> g = Graph[{1 -> 2, 2 -> 3}]; ConnectedGraphQ[g]
     = False

    >> g = Graph[{1 -> 2, 2 -> 3, 3 -> 1}]; ConnectedGraphQ[g]
     = True

    #> g = Graph[{1 -> 2, 2 -> 3, 2 -> 3, 3 -> 1}]; ConnectedGraphQ[g]
     = True

    #> g = Graph[{1 -> 2, 2 -> 3}]; ConnectedGraphQ[g]
     = False

    >> g = Graph[{1 <-> 2, 2 <-> 3}]; ConnectedGraphQ[g]
     = True

    >> g = Graph[{1 <-> 2, 2 <-> 3, 4 <-> 5}]; ConnectedGraphQ[g]
     = False

    #> ConnectedGraphQ[Graph[{}]]
     = True

    #> ConnectedGraphQ["abc"]
     = False
    """

    summary_text = "test if a graph is a connected"

    def eval(self, graph, expression, evaluation, options):
        "%(name)s[graph_, OptionsPattern[%(name)s]]"
        graph = self._build_graph(graph, evaluation, options, expression, quiet=True)
        if graph:
            return from_python(is_connected(graph.G))
        else:
            return SymbolFalse


class DirectedGraphQ(_NetworkXBuiltin):
    """
    <url>
    :Directed graph:
    https://en.wikipedia.org/wiki/Directed_graph
    </url> test (<url>
    :NetworkX:
    https://networkx.org/documentation/networkx-2.8.8/reference\
/generated/networkx.classes.function.is_directed.html
    </url>, <url>
    :WMA:
    https://reference.wolfram.com/language/ref/DirectedGraphQ.html</url>)

    <dl>
      <dt>'DirectedGraphQ'[$graph$]
      <dd>True if $graph$ is a 'Graph' and all the edges are directed.
    </dl>

    >> g = Graph[{1 -> 2, 2 -> 3}]; DirectedGraphQ[g]
     = True

    >> g = Graph[{1 -> 2, 2 <-> 3}]; DirectedGraphQ[g]
     = False

    #> g = Graph[{}]; DirectedGraphQ[{}]
     = False

    #> DirectedGraphQ["abc"]
     = False
    """

    summary_text = "test if a graph is directed"

    def eval(self, graph, expression, evaluation, options):
        "DirectedGraphQ[graph_, OptionsPattern[DirectedGraphQ]]"
        graph = self._build_graph(graph, evaluation, options, expression, quiet=True)
        if graph:
            return from_python(graph.is_directed())
        else:
            return SymbolFalse


class LoopFreeGraphQ(_NetworkXBuiltin):
    """
    <url>
    :Loop-Free graph:
    https://en.wikipedia.org/wiki/Loop_(graph_theory)
    </url> test (<url>
    :NetworkX:
    https://networkx.org/documentation/networkx-2.8.8/reference/\
generated/networkx.classes.function.nodes_with_selfloops.html
    </url>, <url>
    :WMA:
    https://reference.wolfram.com/language/ref/LoopFreeGraphQ.html</url>)

    <dl>
      <dt>'LoopFreeGraphQ'[$graph$]
      <dd>True if $graph$ is a 'Graph' and the edges do not close any loop.
    </dl>

    >> g = Graph[{1 -> 2, 2 -> 3}]; LoopFreeGraphQ[g]
     = True

    >> g = Graph[{1 -> 2, 2 -> 3, 1 -> 1}]; LoopFreeGraphQ[g]
     = False

    #> g = Graph[{}]; LoopFreeGraphQ[{}]
     = False

    #> LoopFreeGraphQ["abc"]
     = False
    """

    summary_text = "test if a graph is loop free"

    def eval(self, graph, expression, evaluation, options):
        "LoopFreeGraphQ[graph_, OptionsPattern[LoopFreeGraphQ]]"
        graph = self._build_graph(graph, evaluation, options, expression, quiet=True)
        if not graph or graph.empty():
            return SymbolFalse

        return from_python(graph.is_loop_free())


class MixedGraphQ(_NetworkXBuiltin):
    """
    <url>
    :Mixed Graph:
    https://en.wikipedia.org/wiki/Mixed_graph</url> test (<url>:WMA:
    https://reference.wolfram.com/language/ref/MixedGraphQ.html</url>)

    <dl>
      <dt>'MixedGraphQ'[$graph$]
      <dd>returns 'True' if $graph$ is a 'Graph' with both directed and undirected edges, \
          and 'False' otherwise.
    </dl>

    >> MixedGraphQ[Graph[{1 -> 2, 2 -> 3}]]
     = False

    >> MixedGraphQ[Graph[{1 -> 2, 2 <-> 3}]]
     = True

    >> MixedGraphQ[Graph[{}]]
     = False

    >> MixedGraphQ["abc"]
     = False

    ## Add as pytests
    ## > g = Graph[{1 -> 2, 2 -> 3}]; MixedGraphQ[g]
    ##  = False
    ## > g = EdgeAdd[g, a <-> b]; MixedGraphQ[g]
    ##   = True
    ## > g = EdgeDelete[g, a <-> b]; MixedGraphQ[g]
    ## = False
    """

    summary_text = "test if a graph has directed and undirected edges"

    def eval(self, graph, expression, evaluation, options):
        "MixedGraphQ[graph_, OptionsPattern[MixedGraphQ]]"
        graph = self._build_graph(graph, evaluation, options, expression, quiet=True)
        if graph:
            return from_python(graph.is_mixed_graph())
        return SymbolFalse


class MultigraphQ(_NetworkXBuiltin):
    """
    <url>
    :Multigraph:
    https://en.wikipedia.org/wiki/Multigraph</url> test (<url>
    :NetworkX:
https://networkx.org/documentation/networkx-2.8.8/reference/classes/multigraph.html</url>, \
    <url>
    :WMA:
    https://reference.wolfram.com/language/ref/MulitGraphQ.html</url>)


    <dl>
      <dt>'MultigraphQ'[$graph$]
      <dd>True if $graph$ is a 'Graph' and there vertices connected by more than \
    one edge.
    </dl>

    >> g = Graph[{1 -> 2, 2 -> 3}]; MultigraphQ[g]
     = False

    >> g = Graph[{1 -> 2, 2 -> 3, 1 -> 2}]; MultigraphQ[g]
     = True

    #> g = Graph[{}]; MultigraphQ[g]
     = False

    #> MultigraphQ["abc"]
     = False
    """

    summary_text = "test if a graph is a multi graph"

    def eval(self, graph, expression, evaluation, options):
        "MultigraphQ[graph_, OptionsPattern[MultigraphQ]]"
        graph = self._build_graph(graph, evaluation, options, expression, quiet=True)
        if graph:
            return from_python(graph.is_multigraph())
        else:
            return SymbolFalse


class PathGraphQ(_NetworkXBuiltin):
    """
    <url>
    :Path graph:
    https://en.wikipedia.org/wiki/Path_graph
    </url> test (<url>
    :WMA:
    https://reference.wolfram.com/language/ref/PathGraphQ.html</url>)

    <dl>
      <dt>'LoopFreeGraphQ'[$graph$]
      <dd>True if $graph$ is a 'Graph' and it becomes disconnected by removing \
    a single edge.
    </dl>


    >> PathGraphQ[Graph[{1 -> 2, 2 -> 3}]]
     = True
    #> PathGraphQ[Graph[{1 -> 2, 2 -> 3, 3 -> 1}]]
     = True
    #> PathGraphQ[Graph[{1 <-> 2, 2 <-> 3}]]
     = True
    >> PathGraphQ[Graph[{1 -> 2, 2 <-> 3}]]
     = False
    >> PathGraphQ[Graph[{1 -> 2, 3 -> 2}]]
     = False
    >> PathGraphQ[Graph[{1 -> 2, 2 -> 3, 2 -> 4}]]
     = False
    >> PathGraphQ[Graph[{1 -> 2, 3 -> 2, 2 -> 4}]]
     = False

    #> PathGraphQ[Graph[{}]]
     = False
    #> PathGraphQ[Graph[{1 -> 2, 3 -> 4}]]
     = False
    #> PathGraphQ[Graph[{1 -> 2, 2 -> 1}]]
     = True
    >> PathGraphQ[Graph[{1 -> 2, 2 -> 3, 2 -> 3}]]
     = False
    #> PathGraphQ[Graph[{}]]
     = False
    #> PathGraphQ["abc"]
     = False
    #> PathGraphQ[{1 -> 2, 2 -> 3}]
     = False
    """

    summary_text = "test if a graph is a path-like graph"

    def eval(self, graph, expression, evaluation, options):
        "PathGraphQ[graph_, OptionsPattern[PathGraphQ]]"
        if not isinstance(graph, Graph) or graph.empty():
            return SymbolFalse

        G = graph.G

        if G.is_directed():
            connected = nx.is_semiconnected(G)
        else:
            connected = nx.is_connected(G)

        if connected:
            is_path = all(d <= 2 for _, d in G.degree(graph.vertices))
        else:
            is_path = False

        return from_python(is_path)


class PlanarGraphQ(_NetworkXBuiltin):
    """
        <url>
        :Planar Graph:
        https://en.wikipedia.org/wiki/Planar_graph</url> test (<url>
        :NetworkX:
        https://networkx.org/documentation/networkx-2.8.8/reference/algorithms/
    generated/networkx.algorithms.planarity.check_planarity.html</url>, <url>
        :WMA:
        https://reference.wolfram.com/language/ref/PlanaGraphQ.html</url>)


        <dl>
          <dt>'PlanarGraphQ'[$g$]
          <dd>Returns True if $g$ is a planar graph and False otherwise.
        </dl>

        >> PlanarGraphQ[CycleGraph[4]]
        = True

        >> PlanarGraphQ[CompleteGraph[5]]
        = False

        #> PlanarGraphQ[Graph[{}]]
         = False


        >> PlanarGraphQ["abc"]
         : Expected a graph at position 1 in PlanarGraphQ[abc].
         = False
    """

    options = DEFAULT_GRAPH_OPTIONS
    summary_text = "test if a graph is planar"

    def eval(self, graph, expression, evaluation: Evaluation, options: dict):
        "Pymathics`PlanarGraphQ[graph_, OptionsPattern[PlanarGraphQ]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if not graph or graph.empty():
            return SymbolFalse
        is_planar, _ = nx.check_planarity(graph.G)
        return from_python(is_planar)


class SimpleGraphQ(_NetworkXBuiltin):
    """
    Simple (not multigraph) <url>
    :graph:
    https://en.wikipedia.org/wiki/Graph_(discrete_mathematics)#Graph
    </url> test (<url>
    :WMA:
    https://reference.wolfram.com/language/ref/SimpleGraphQ.html</url>)

    <dl>
      <dt>'SimpleGraphQ'[$graph$]
      <dd>True if $graph$ is a 'Graph', loop-free and each pair of \
          vertices are connected at most by a single edge.
    </dl>

    >> g = Graph[{1 -> 2, 2 -> 3, 3 <-> 4}]; SimpleGraphQ[g]
     = True

    >> g = Graph[{1 -> 2, 2 -> 3, 1 -> 1}]; SimpleGraphQ[g]
     = False

    >> g = Graph[{1 -> 2, 2 -> 3, 1 -> 2}]; SimpleGraphQ[g]
     = False

    #> SimpleGraphQ[Graph[{}]]
     = True

    #> SimpleGraphQ["abc"]
     = False
    """

    summary_text = "test if a graph is simple (not multigraph)"

    def eval(self, graph, expression, evaluation, options):
        "SimpleGraphQ[graph_, OptionsPattern[LoopFreeGraphQ]]"
        graph = self._build_graph(graph, evaluation, options, expression, quiet=True)
        if graph:
            if graph.empty():
                return SymbolTrue
            else:
                simple = graph.is_loop_free() and not graph.is_multigraph()
                return from_python(simple)
        else:
            return SymbolFalse



1  # -*- coding: utf-8 -*-

"""
Low level functions to format graphs
"""

import math
import networkx as nx
import random

from io import BytesIO
from matplotlib import pyplot

cached_pair = None
node_size = 300  # this is networkx's default size

DEFAULT_NODE_SIZE = 300.0
DEFAULT_POINT_SIZE = 16
LAYOUT_DENSITY_EXPONENT = {"circular": 0.9, "spiral_equidistant": 0.7, "spiral": 0.6}


def clamp(value, min=-math.inf, max=math.inf):
    if value <= min:
        return min
    if value >= max:
        return max
    return value


def format_graph(G, **options) -> str:
    """
    Format a networkx graph using nx.draw (using matplotlib) and
    return a SVG string that encodes the graph.
    """

    global node_size
    global cached_pair
    pyplot.close()
    pyplot.switch_backend("AGG")
    cached_pair = None

    graph_layout = G.graph_layout if hasattr(G, "graph_layout") else None
    node_shape = G.node_shape if hasattr(G, "node_shape") else "o"

    node_size = DEFAULT_NODE_SIZE
    draw_options = {
        "node_size": node_size,
        "node_shape": node_shape,
        # "with_labels": vertex_labels # Set below
        # "font_size": 12,        # Harmonized
        # "node_color": "white",  # Set below
        # "edgecolors": "black",  # Set below
        # "width": 5,             # Marmonized
    }
    draw_options.update(options)

    vertex_labels = G.vertex_labels if hasattr(G, "vertex_labels") else False
    if vertex_labels:
        draw_options["with_labels"] = bool(vertex_labels)

    if hasattr(G, "title") and G.title:
        fig, ax = pyplot.subplots()  # Create a figure and an axes
        ax.set_title(G.title)

    layout_fn = None
    if graph_layout:
        if not isinstance(graph_layout, str):
            graph_layout = graph_layout.get_string_value()
        layout_fn = NETWORKX_LAYOUTS.get(graph_layout, None)
        if graph_layout in ["circular", "spiral", "spiral_equidistant"]:
            pyplot.axes().set_aspect("equal")

    harmonize_parameters(G, draw_options)

    if layout_fn:
        nx.draw(G, pos=layout_fn(G), **draw_options)
    else:
        nx.draw_shell(G, **draw_options)


def get_svg_graph() -> str:
    """
    Retrieves SVG XML from what has already been stored the matplotlib buffer
    pyplot. Return the XML SVG string.
    """

    buffer = BytesIO()
    pyplot.savefig(buffer, format="svg")
    buffer.seek(0)
    graph_svg = buffer.getvalue()
    buffer.close()

    # TODO: In the future if probably want to base64 encode.
    # Use: base64.b64encode(some peice of image_svg)
    return graph_svg.decode("utf-8")


def get_png_graph() -> BytesIO:
    """
    Retrieves SVG XML from what has already been stored the matplotlib buffer
    pyplot. Return the XML SVG string.
    """

    buffer = BytesIO()
    pyplot.savefig(buffer, format="png")
    buffer.seek(0)
    graph_png = buffer.getvalue()
    buffer.close()

    # TODO: In the future if probably want to base64 encode.
    # Use: base64.b64encode(some peice of image_svg)
    return graph_png


def hierarchy_pos(
    G, root=None, width=1.0, vert_gap=0.2, vert_loc=0, leaf_vs_root_factor=0.5
):

    """Position nodes in tree layout. The root is at the top.

    Based on Joel's answer at https://stackoverflow.com/a/29597209/2966723,
    but with some modifications.

    There are two basic approaches we think of to allocate the horizontal
    location of a node.

    - Top down: we allocate horizontal space to a node.  Then its ``k``
      descendants split up that horizontal space equally.  This tends to result
      in overlapping nodes when some have many descendants.
    - Bottom up: we allocate horizontal space to each leaf node.  A node at a
      higher level gets the entire space allocated to its descendant leaves.
      Based on this, leaf nodes at higher levels get the same space as leaf
      nodes very deep in the tree.

    We use use both of these approaches simultaneously with ``leaf_vs_root_factor``
    determining how much of the horizontal space is based on the bottom up
    or top down approaches.  ``0`` gives pure bottom up, while 1 gives pure top
    down.

    From EoN (Epidemics on Networks): a fast, flexible Python package
    for simulation, analytic approximation, and analysis of epidemics
    on networks
    https://joss.theoj.org/papers/10.21105/joss.01731

    :Arguments:

    Parameters
    ----------
    G : NetworkX graph or list of nodes
        A position will be assigned to every node in G.
        The graph must be a tree.

    root : the root node of the tree

    - if the tree is directed and this is not given, the root will be found and used
    - if the tree is directed and this is given, then the positions will be
      just for the descendants of this node.
    - if the tree is undirected and not given, then a random choice will be used.

    width : horizontal space allocated for this branch - avoids overlap with other branches

    vert_gap : gap between levels of hierarchy

    vert_loc : vertical location of root

    leaf_vs_root_factor : used in calculating the _x_ coordinate of a leaf

    xcenter : horizontal location of root

    Examples
    --------
    >>> G = nx.binomial_tree(3)
    >>> nx.draw(G, pos=nx.hierarchy_layout(G, root=0))

    As the number of nodes gets large, the node size and node labels
    may need to be adjusted. The following shows how the minimum
    separation between nodes can be used to adjust node sizes.

    >>> G = nx.full_rary_tree(2, 127)
    >>> pos, min_sep = nx.hierarchy_layout_with_min_sep(G, root=0)
    >>> nx.draw(G, pos=pos, node_size=min_sep * 1500)

    Also see the NetworkX drawing examples at
    https://networkx.org/documentation/latest/auto_examples/index.html

    """
    if not nx.is_tree(G):
        raise TypeError("cannot use hierarchy_pos on a graph that is not a tree")

    global cached_pair
    if cached_pair is not None:
        return cached_pair

    # These get swapped if tree edge directions point to the root.
    decendants = nx.descendants
    out_degree = G.out_degree if hasattr(G, "out_degree") else G.degree
    neighbors = G.neighbors

    if root is None:
        if isinstance(G, nx.DiGraph):
            zero_outs = [n for n in G.out_degree() if n[1] == 0]
            if len(zero_outs) == 1 and len(G) > 2:
                # We unequivocally have a directed that points from leave to the root.
                # The case where we have a one or two node graph is ambiguous.
                root = list(nx.topological_sort(G))[-1]
                # Swap motion functions
                decendants = nx.ancestors
                out_degree = G.in_degree
                neighbors = G.predecessors
            else:
                root = next(
                    iter(nx.topological_sort(G))
                )  # allows back compatibility with nx version 1.11
                # root = next(nx.topological_sort(G))
        else:
            root = random.choice(list(G.nodes))

    def _hierarchy_pos(
        G,
        root,
        leftmost,
        width,
        leafdx=0.2,
        vert_gap=0.2,
        vert_loc=0,
        xcenter=0.5,
        rootpos=None,
        leafpos=None,
        parent=None,
    ):
        """
        see hierarchy_pos docstring for most arguments

        pos: a dict saying where all nodes go if they have been assigned
        parent: parent of this branch. - only affects it if non-directed

        """

        if rootpos is None:
            rootpos = {root: (xcenter, vert_loc)}
        else:
            rootpos[root] = (xcenter, vert_loc)
        if leafpos is None:
            leafpos = {}

        children = list(neighbors(root))
        leaf_count = 0
        if not isinstance(G, nx.DiGraph) and parent is not None:
            children.remove(parent)
        if len(children) != 0:
            rootdx = width / len(children)
            nextx = xcenter - width / 2 - rootdx / 2
            for child in children:
                nextx += rootdx
                rootpos, leafpos, newleaves = _hierarchy_pos(
                    G,
                    child,
                    leftmost + leaf_count * leafdx,
                    width=rootdx,
                    leafdx=leafdx,
                    vert_gap=vert_gap,
                    vert_loc=vert_loc - vert_gap,
                    xcenter=nextx,
                    rootpos=rootpos,
                    leafpos=leafpos,
                    parent=root,
                )
                leaf_count += newleaves

            leftmostchild = min((x for x, y in [leafpos[child] for child in children]))
            rightmostchild = max((x for x, y in [leafpos[child] for child in children]))
            leafpos[root] = ((leftmostchild + rightmostchild) / 2, vert_loc)
        else:
            leaf_count = 1
            leafpos[root] = (leftmost, vert_loc)
        #        pos[root] = (leftmost + (leaf_count-1)*dx/2., vert_loc)
        #        print(leaf_count)
        return rootpos, leafpos, leaf_count

    xcenter = width / 2.0
    if isinstance(G, nx.DiGraph):
        leafcount = len([node for node in decendants(G, root) if out_degree(node) == 0])
    elif isinstance(G, nx.Graph):
        leafcount = len(
            [
                node
                for node in nx.node_connected_component(G, root)
                if G.degree(node) == 1 and node != root
            ]
        )

    rootpos, leafpos, leaf_count = _hierarchy_pos(
        G,
        root,
        0,
        width,
        leafdx=width * 1.0 / leafcount,
        vert_gap=vert_gap,
        vert_loc=vert_loc,
        xcenter=xcenter,
    )
    pos = {}
    for node in rootpos:
        pos[node] = (
            leaf_vs_root_factor * leafpos[node][0]
            + (1 - leaf_vs_root_factor) * rootpos[node][0],
            leafpos[node][1],
        )
    #    pos = {node:(leaf_vs_root_factor*x1+(1-leaf_vs_root_factor)*x2, y1) for ((x1,y1), (x2,y2)) in (leafpos[node], rootpos[node]) for node in rootpos}
    xmax = max(x for x, y in pos.values())
    y_list = {}
    for node in pos:
        x, y = pos[node] = (pos[node][0] * width / xmax, pos[node][1])
        y_list[y] = y_list.get(y, set([]))
        y_list[y].add(x)

    min_sep = xmax
    for y in y_list.keys():
        x_list = sorted(y_list[y])
        n = len(x_list) - 1
        if n <= 0:
            continue
        min_sep = min([x_list[i + 1] - x_list[i] for i in range(n)] + [min_sep])
    cached_pair = pos, min_sep
    return cached_pair


def harmonize_parameters(G, draw_options: dict):
    global node_size
    graph_layout = G.graph_layout if hasattr(G, "graph_layout") else ""

    if graph_layout == "tree":
        # Call this to compute node_size. Cache the
        # results
        tree_layout(G)
        draw_options["node_size"] = node_size
    elif graph_layout in ["circular", "spiral", "spiral_equidistant"]:
        exponent = LAYOUT_DENSITY_EXPONENT[graph_layout]
        node_size = draw_options["node_size"] = (2 * DEFAULT_NODE_SIZE) / (
            len(G) + 1
        ) ** exponent
        # print("XX", node_size, exponent)

    if draw_options.get("with_labels", False):
        draw_options["edgecolors"] = draw_options.get("edgecolors", "black")
        draw_options["node_color"] = draw_options.get("node_color", "white")

    if "width" not in draw_options:
        width = clamp(node_size / DEFAULT_NODE_SIZE, min=0.15)
        draw_options["width"] = width

    if "font_size" not in draw_options:
        # FIXME: should also take into consideration max width of label.
        font_size = clamp(
            int((node_size * DEFAULT_POINT_SIZE) / DEFAULT_NODE_SIZE), min=1, max=18
        )
        draw_options["font_size"] = font_size


def svg_format_graph(G, **options):
    format_graph(G, **options)
    # pyplot.tight_layout()
    svg_graph_xml = get_svg_graph()
    svg_str = svg_graph_xml[svg_graph_xml.find("<svg xmlns:xlink") :]
    return svg_str


def png_format_graph(G, **options):
    format_graph(G, **options)
    # pyplot.tight_layout()
    return get_png_graph()


def spiral_equidistant_layout(G, *args, **kwargs):
    return nx.spiral_layout(G, equidistant=True, *args, **kwargs)


def tree_layout(G):
    global node_size
    root = G.root if hasattr(G, "root") else None
    pos, min_sep = hierarchy_pos(G, root=root)
    node_size = min_sep * 2000
    return pos


NETWORKX_LAYOUTS = {
    "circular": nx.circular_layout,
    "kamada_kawai": nx.kamada_kawai_layout,
    "multipartite": nx.multipartite_layout,
    "planar": nx.planar_layout,
    "random": nx.random_layout,
    "shell": nx.shell_layout,
    "spectral": nx.spectral_layout,
    "spiral": nx.spiral_layout,
    "spiral_equidistant": spiral_equidistant_layout,
    "spring": nx.spring_layout,
    "tree": tree_layout,
}



"""
Graph Operations and Modifications
"""

import networkx as nx
from mathics.core.convert.python import from_python
from mathics.core.evaluation import Evaluation
from mathics.core.expression import Expression

from pymathics.graph.base import (
    DEFAULT_GRAPH_OPTIONS,
    SymbolDirectedEdge,
    SymbolUndirectedEdge,
    _create_graph,
    _NetworkXBuiltin,
)


class FindSpanningTree(_NetworkXBuiltin):
    """
    <url>:Spanning Tree:https://en.wikipedia.org/wiki/Spanning_tree</url> \
    (<url>:NetworkX:
    https://networkx.org/documentation/networkx-2.8.8/reference/algorithms/\
generated/networkx.algorithms.tree.mst.minimum_spanning_edges.html
    </url>,\
    <url>:WMA:
    https://reference.wolfram.com/language/ref/FindSpanningTree.html
    </url>)

    <dl>
      <dt>'FindSpanningTree'[$g$]
      <dd>finds a spanning tree of the graph $g$.
    </dl>

    >> FindSpanningTree[CycleGraph[4]]
     = -Graph-
    """

    options = DEFAULT_GRAPH_OPTIONS
    summary_text = "find a spanning tree"

    def eval(self, graph, expression, evaluation: Evaluation, options: dict):
        "%(name)s[graph_, OptionsPattern[%(name)s]]"
        graph = self._build_graph(graph, evaluation, options, expression)
        if graph:
            graph.update_weights(evaluation)
            SymbolDirectedEdge if graph.G.is_directed() else SymbolUndirectedEdge
            # FIXME: put in edge to Graph conversion function?
            edges = [
                Expression(SymbolUndirectedEdge, from_python(u), from_python(v))
                for u, v in nx.minimum_spanning_edges(graph.G, data=False)
            ]
            g = _create_graph(edges, [None] * len(edges), options)
            if not hasattr(g.G, "graph_layout"):
                if hasattr(graph.G, "graph_layout"):
                    g.G.graph_layout = graph.G.graph_layout
                else:
                    g.G.graph_layout = "tree"
            return g


# TODO: Subgraph, NeighborhoodGraph + many others



# -*- coding: utf-8 -*-
"""
Unit tests for pymathics.graph.algorithms
"""
from test.helper import check_evaluation, evaluate, evaluate_value
import pytest

def setup_module(module):
    """Load pymathics.graph"""
    assert evaluate_value('LoadModule["pymathics.graph"]') == "pymathics.graph"
    evaluate("SortList[list_] := Sort[Map[Sort, list]]")



@pytest.mark.skip("Wrong result. Investigate me.")
def test_EdgeList():
    check_evaluation(
     "EdgeList[{1 -> 2, 2 <-> 3}]",
     "{DirectedEdge[1, 2], UndirectedEdge[2, 3]}"
     )


    
@pytest.mark.skip("Wrong result. Investigate me.")
def test_FindShortestPath():
     check_evaluation(
         (
             "g = Graph[{1 -> 2, 2 -> 3, 1 -> 3}, EdgeWeight -> {0.5, a, 3}];"
             "a = 0.5; FindShortestPath[g, 1, 3]'"),
         "{1, 2, 3}")
     check_evaluation("a = 10; FindShortestPath[g, 1, 3]", "{1, 3}")
     check_evaluation("a = .;", "Null")




@pytest.mark.skip("This finds d<->a in the position 4 instead 2.")
def test_EdgeIndex():
    check_evaluation("EdgeIndex[{c <-> d, d <-> a, a -> e}, d <-> a]",
                     "2")

    
@pytest.mark.parametrize(
    ("str_expr", "str_expect"),
    [
        (
            ("g = Graph[{1 -> 2, 2 -> 3}, DirectedEdges -> True];"
             "EdgeCount[g, _DirectedEdge]"), 
            "2"
        ),
        (
            (
                "g = Graph[{1 -> 2, 2 -> 3}, DirectedEdges -> False];"
                "EdgeCount[g, _DirectedEdge]"),
            "0"
        ),
        ("EdgeCount[g, _UndirectedEdge]","2")
    ]
)
@pytest.mark.skip("This finds d<->a in the position 4 instead 2.")
def test_edgecount(str_expr, str_expect):
    check_evaluation(str_expr, str_expect)

    
@pytest.mark.skip("This finds d<->a in the position 4 instead 2.")
def test_EdgeIndex():
    check_evaluation("EdgeIndex[{c <-> d, d <-> a, a -> e}, d <-> a]",
                     "2")

@pytest.mark.skip("This is not properly evaluated. Investigate me")
def test_HITSCentrality():
     check_evaluation("g = Graph[{a -> d, b -> c, d -> c, d -> a, e -> c}]; HITSCentrality[g]",
     "{{0.292893, 0., 0., 0.707107, 0.}, {0., 1., 0.707107, 0., 0.707107}}")

@pytest.mark.skip("Investigate me.")
def test_EdgeRules():
    check_evaluation(
        "EdgeRules[{1 <-> 2, 2 -> 3, 3 <-> 4}]",
     "{1 -> 2, 2 -> 3, 3 -> 4}"
    )
     



# -*- coding: utf-8 -*-
"""
Unit tests for pymathics.graph.base
"""
from test.helper import check_evaluation, evaluate, evaluate_value

import pytest


def setup_module(module):
    """Load pymathics.graph"""
    assert evaluate_value('LoadModule["pymathics.graph"]') == "pymathics.graph"
    evaluate("SortList[list_] := Sort[Map[Sort, list]]")


@pytest.mark.parametrize(
    ("str_expr", "str_expected", "msg"),
    [
        (None, None, None),
        (
            "g = Graph[{a -> b, b <-> c, d -> c, d -> a, e <-> c, d -> b}];",
            "Null",
            "Intialize graph",
        ),
        ("Sort[DegreeCentrality[g]]", "{2, 2, 3, 4, 5}", None),
        ('Sort[DegreeCentrality[g, "In"]]', "{0, 1, 1, 3, 3}", None),
        ('Sort[DegreeCentrality[g, "Out"]]', "{1, 1, 1, 2, 3}", None),
        (None, None, None),
    ],
)
def test_degree_centrality(str_expr, str_expected, msg):
    check_evaluation(str_expr, str_expected, failure_message=msg)


@pytest.mark.parametrize(
    ("str_expr", "str_expected", "msg"),
    [
        (None, None, None),
        (
            "Length[EdgeList[EdgeDelete[{a -> b, b -> c, c -> d}, b -> c]]]",
            "2",
            None,
        ),
        # ("Length[EdgeList[EdgeDelete[{a -> b, b -> c, c -> b, c -> d}, b <-> c]]]", "4", None),
        ("Length[EdgeList[EdgeDelete[{a -> b, b <-> c, c -> d}, b -> c]]]", "3", None),
        ("Length[EdgeList[EdgeDelete[{a -> b, b <-> c, c -> d}, c -> b]]]", "3", None),
        ("Length[EdgeList[EdgeDelete[{a -> b, b <-> c, c -> d}, b <-> c]]]", "2", None),
        (None, None, None),
    ],
)
def test_edge_delete(str_expr, str_expected, msg):
    check_evaluation(str_expr, str_expected, failure_message=msg)



# -*- coding: utf-8 -*-
"""
Unit tests for pymathics.graph.tree
"""
from test.helper import check_evaluation, evaluate_value


def setup_module(module):
    """Load pymathics.graph"""
    assert evaluate_value('LoadModule["pymathics.graph"]') == "pymathics.graph"


def test_tree():
    for str_expr, str_expected in [
        ("TreeGraphQ[StarGraph[3]]", "True"),
        ("TreeGraphQ[CompleteGraph[0]]", "False"),
        ("TreeGraphQ[CompleteGraph[1]]", "True"),
        ("TreeGraphQ[CompleteGraph[2]]", "True"),
        ("TreeGraphQ[CompleteGraph[3]]", "False"),
    ]:
        check_evaluation(str_expr, str_expected)



import networkx as nx
from mathics.core.evaluation import Evaluation
from mathics.core.expression import Expression

from pymathics.graph.base import (
    Graph,
    SymbolUndirectedEdge,
    _graph_from_list,
    _NetworkXBuiltin,
)
from pymathics.graph.tree import DEFAULT_TREE_OPTIONS


class PathGraph(_NetworkXBuiltin):
    """
    <url>
    :WMA:https://reference.wolfram.com/language/ref/PathGraph.html
    </url>

    <dl>
      <dt>'PathGraph[{$v_1$, $v_2$, ...}]'
      <dd>Returns a Graph with a path with vertices $v_i$ and edges between $v-i$ and $v_i+1$ .
    </dl>

    >> PathGraph[{1, 2, 3}]
     = -Graph-
    """

    summary_text = "create a graph from a path"

    def eval(self, e, evaluation: Evaluation, options: dict) -> Graph:
        "PathGraph[e_List, OptionsPattern[PathGraph]]"
        elements = e.elements

        def edges():
            for u, v in zip(elements, elements[1:]):
                yield Expression(SymbolUndirectedEdge, u, v)

        g = _graph_from_list(edges(), options)
        g.G.graph_layout = (
            options["System`GraphLayout"].get_string_value() or "spiral_equidistant"
        )
        return g


class TreeGraph(Graph):
    """
    <url>
    :WMA:https://reference.wolfram.com/language/ref/TreeGraph.html
    </url>

    <dl>
      <dt>'TreeGraph[{$edge_1$, $edge_2$, ...}]'
      <dd>create a tree-like from a list of edges.
    </dl>


    >> TreeGraph[{1->2, 2->3, 2->4}]
     = -Graph-

    """

    options = DEFAULT_TREE_OPTIONS

    messages = {
        "notree": "Graph is not a tree.",
    }

    summary_text = "build a tree graph"

    def __init__(self, G, **kwargs):
        super(Graph, self).__init__()
        if not nx.is_tree(G):
            raise ValueError
        self.G = G


# TODO: PlanarGraph



# -*- coding: utf-8 -*-
"""
Unit tests for pymathics.graph.parametric
"""
from test.helper import check_evaluation, evaluate, evaluate_value


def setup_module(module):
    """Load pymathics.graph"""
    assert evaluate_value('LoadModule["pymathics.graph"]') == "pymathics.graph"
    evaluate("SortList[list_] := Sort[Map[Sort, list]]")


def test_complete_graph():
    for str_expr, str_expected, mess in [
        (
            "CompleteGraph[0]",
            "CompleteGraph[0]",
            ["Expected a positive integer at position 1 in CompleteGraph[0]."],
        ),
    ]:
        check_evaluation(str_expr, str_expected, expected_messages=mess)
